
/*****************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 2001 Reid Simmons.  All rights reserved.
 *
 * FILE: tcmThread.h
 *
 * ABSTRACT: Encapsulates basic functional interface to threading sub-system.
 *           Used both with and without THREADED being defined.
 *
 * $Source: /afs/cs/project/TCA/Master/tcaV8/tcm/src/tcmThread.h,v $ 
 * $Revision: 1.1 $
 * $Date: 2001/10/23 22:53:00 $
 * $Author: da0g $
 *
 * REVISION HISTORY:
 *
 * $Log: tcmThread.h,v $
 * Revision 1.1  2001/10/23 22:53:00  da0g
 * Added Threading support.  Cached lastchild.
 *
 *
 *****************************************************************************/

#ifndef TCM_THREAD_H
#define TCM_THREAD_H

#ifdef THREADED
#include "ThreadManager.h"
#endif /* THREADED */


inline void ExecuteTask ( const Task_Tree_Ref & theTask )
{
  if ( validRef ( theTask ) )
  {


#ifdef THREADED
    if ( theTask -> getIsThreaded() )
    {
      ThreadManager::getGlobalThreadManager().executeTask ( theTask );
    }
    else
    {
      	/* If we are threaded, fully unlock ourself before waiting... */
      unsigned int numberOfLocks
	= TCM_FullyUnlockMasterMutexReturningNumberOfLocks();
#endif /* THREADED */



	/* Run Task in non-threaded mode. */
      theTask -> execute();



#ifdef THREADED
	/* If we are threaded, fully re-lock ourself afterwards... */
      TCM_FullyRelockMasterMutexNumberOfLocksTimes ( numberOfLocks );
    }
#endif /* THREADED */


  }
  else
  {
    tcmWarning("ExecuteTask: Reference to NULL node\n");
  }
}


inline BOOLEAN SetPrimaryThread (
	    BOOLEAN   theReportError                       = TRUE,
	    BOOLEAN   theReturnValueIfAlreadyPrimaryThread = TRUE,
	    BOOLEAN   theReportIfAlreadyPrimaryThread      = TRUE,
	    BOOLEAN * theWasAlreadyPrimaryThread           = (BOOLEAN *) NULL )
{
  BOOLEAN returnValue = TRUE;

#ifdef THREADED
  returnValue = TO_BOOLEAN (
       ThreadManager::getGlobalThreadManager() . setPrimaryThread (
	 theReportError,
	 ((theReturnValueIfAlreadyPrimaryThread == TRUE) ? SUCCESS : FAILURE),
	 theReportIfAlreadyPrimaryThread,
	 theWasAlreadyPrimaryThread )
    == SUCCESS );
#endif /* THREADED */  

  return returnValue;
}


inline BOOLEAN ClearPrimaryThread()
{
  BOOLEAN returnValue = TRUE;

#ifdef THREADED
  returnValue = TO_BOOLEAN (    ThreadManager::getGlobalThreadManager()
				  . clearPrimaryThread()
			     == SUCCESS );
#endif /* THREADED */  

  return returnValue;
}


#ifdef THREADED
inline THREAD_ID GetPrimaryThreadId()
{
  return ThreadManager::getGlobalThreadManager().getPrimaryThreadId();
}
#endif


inline BOOLEAN HasThreadWithUnfinishedWork()
{
  BOOLEAN returnValue = TRUE;

#ifdef THREADED
  returnValue
    = ThreadManager::getGlobalThreadManager() . hasThreadWithUnfinishedWork();
#endif /* THREADED */  

  return returnValue;
}

#endif /* TCM_THREAD_H */
