
import netscape.application.Color;
import netscape.util.*;


/*
 * A Compound Statement is basically "{" 0-or-more-statements "}".
 *
 * (But it forms the basis for multiple statements in C/C++/TDL programs.)
 */

public class DataCompoundStatement extends DataStatement
{
	/* Class Constants */
  public final static String   OPEN_BRACE     = "{";
  public final static String   CLOSE_BRACE    = "}";
  public final static String   NEW_LINE       = "\n";

  public final static Color    COMPOUND_STATEMENT_COLOR = Color.pink;

	/* Instance Variables */
  protected Vector  modulesVector;/*scratch vector used for calculations only*/
  protected int     lineNumberOfCloseBrace;


	/* Instance Methods */
  public DataCompoundStatement ( Object theComponent )
  {
    this();

	/* Add the open-brace */
    addSubcomponent ( DataCompoundStatement.NEW_LINE );
    setIndex ( DataCompoundStatement.OPEN_BRACE );

	/* Add the statement... */
    if ( theComponent != null )
    {
      if ( theComponent instanceof String )
	addSubcomponent ( (String) theComponent );

      else if ( theComponent instanceof DataComponent )
	addSubcomponent ( (DataComponent) theComponent );

      else
      {
	System.err.println ( "[DataCompoundStatement:DataCompoundStatement]  "
		     + "Warning:  theComponent has an UNACCEPTABLE class (\""
		     + theComponent . getClass() . getName() + "\")." );
      }
    }

	/* Add the close-brace */
    addSubcomponent ( DataCompoundStatement.NEW_LINE );

	/* Note our new close-brace location... */
    setIndex ( DataCompoundStatement.CLOSE_BRACE );
  }

  public DataCompoundStatement()
  {
    modulesVector          = new Vector();
    lineNumberOfCloseBrace = DataComponent.INVALID_LINE_NUMBER;
  }

  public Vector getModulesVector()  { return modulesVector; }

  public GraphModule getModuleFor ( DataComponent  theDataComponent )
  {
    for ( int i=0;  i < getModulesVector() . count();  i++ )
    {
      if ( ( (GraphModule) ( getModulesVector() . elementAt ( i ) ) )
	   . getDataComponent() == theDataComponent )
      {
	return ( (GraphModule) ( getModulesVector() . elementAt ( i ) ) );
      }
    }
    return null;
  }

  public void setLineNumberOfCloseBrace ( int theLineNumberOfCloseBrace )
		   { lineNumberOfCloseBrace = theLineNumberOfCloseBrace; }

  public int  getLineNumberOfCloseBrace() { return lineNumberOfCloseBrace; }

  public boolean hasValidLineNumberOfCloseBrace()
  {
    return DataComponent.isValidLineNumber ( getLineNumberOfCloseBrace() );
  }



  public void addStringAfterOpenBrace ( String theStringToAdd )
  {
    addSubcomponent ( theStringToAdd,
		      getIndex ( DataCompoundStatement.OPEN_BRACE ) );

    setIndex (            DataCompoundStatement.CLOSE_BRACE,
	       getIndex ( DataCompoundStatement.CLOSE_BRACE ) + 1 );
  }

  public void addSubcomponentAfterOpenBrace ( DataComponent theDataComponent )
  {
    addSubcomponent ( theDataComponent,
		      getIndex ( DataCompoundStatement.OPEN_BRACE ) );

    setIndex (            DataCompoundStatement.CLOSE_BRACE,
	       getIndex ( DataCompoundStatement.CLOSE_BRACE ) + 1 );
  }

  public void removeStringAfterOpenBrace ( )
  {
    setIndex (            DataCompoundStatement.CLOSE_BRACE,
	       getIndex ( DataCompoundStatement.CLOSE_BRACE ) - 1 );

    removeSubcomponent ( getIndex ( DataCompoundStatement.OPEN_BRACE ) );
  }

  public void removeSubcomponentAfterOpenBrace ( )
  {
    removeStringAfterOpenBrace();
  }

  public void addStringBeforeCloseBrace ( String theStringToAdd )
  {
    addSubcomponent ( theStringToAdd,
		      getIndex ( DataCompoundStatement.CLOSE_BRACE ) );

    setIndex (            DataCompoundStatement.CLOSE_BRACE,
	       getIndex ( DataCompoundStatement.CLOSE_BRACE ) + 1 );
  }

  public void addSubcomponentBeforeCloseBrace( DataComponent theDataComponent )
  {
    addSubcomponent ( theDataComponent,
		      getIndex ( DataCompoundStatement.CLOSE_BRACE ) );

    setIndex (            DataCompoundStatement.CLOSE_BRACE,
	       getIndex ( DataCompoundStatement.CLOSE_BRACE ) + 1 );
  }

  public void removeStringBeforeCloseBrace ( )
  {
    setIndex (            DataCompoundStatement.CLOSE_BRACE,
	       getIndex ( DataCompoundStatement.CLOSE_BRACE ) - 1 );

    removeSubcomponent ( getIndex ( DataCompoundStatement.CLOSE_BRACE ) );
  }

  public void removeSubcomponentBeforeCloseBrace ( )
  {
    removeStringBeforeCloseBrace();
  }


  public int getChildStatementCount()
  {
    int i, count=0;

    for ( i=0;  i < getSubcomponentsCount();  i++ )
    {
      if ( isSubcomponentADataStatement ( i ) )
      {
	count++;
      }
    }

    return super . getChildStatementCount()  +  count;
  }

  public DataStatement  getChildStatement ( int theIndex )
  {
    int i, count=0;

    for ( i=0;  i < getSubcomponentsCount();  i++ )
    {
      if ( isSubcomponentADataStatement ( i ) )
      {
	if ( theIndex == count )
	  return getDataStatementSubcomponent ( i );
	else
	  count++;
      }
    }

    return super . getChildStatement ( theIndex - count );
  }


  public boolean addPrimaryChild ( DataComponent theChildToAdd,
				   DataComponent theAddChildAfterThisComponent)
  {
    return addChild ( theChildToAdd, theAddChildAfterThisComponent );
  }

  public boolean addChild ( DataComponent theChildToAdd,
			    DataComponent theAddChildAfterThisComponent )
  {
    int i;

	/* Find where to add the child */
    if ( theAddChildAfterThisComponent == null )
    {
      i = getIndex ( DataCompoundStatement.OPEN_BRACE );
    }
    else
    {
      for ( i = getIndex ( DataCompoundStatement.OPEN_BRACE );
	    i < getSubcomponentsCount();
	    i++ )
      {
	if (   (    isSubcomponentADataComponent ( i ) )
	    && (    getDataComponentSubcomponent ( i )
	         == theAddChildAfterThisComponent      ) )
	{
	  break;
	}
      }

      i++; /* Add after this component */
    }

	/* Add the child */
    addSubcomponent ( theChildToAdd, i );

	/* Adjust child's parent */
    theChildToAdd . setParent ( this );

	/* Tweak the } index */
    setIndex ( DataCompoundStatement.CLOSE_BRACE,
	       getIndex ( DataCompoundStatement.CLOSE_BRACE ) + 1 );

    return true;
  }


  public boolean removeChild ( DataComponent theChildToRemove )
  {
    int   index = getIndexOfSubcomponent ( theChildToRemove );

    if (   ( index != DataComponent.INVALID_INDEX )
	&& ( removeSubcomponent ( index )         ) )
    {
	/* Adjust child's parent */
      theChildToRemove . setParent ( null );

	/* Tweak the } index */
      setIndex ( DataCompoundStatement.CLOSE_BRACE,
		 getIndex ( DataCompoundStatement.CLOSE_BRACE ) - 1 );

      return true;
    }
    else
    {
      System.err.println ( "[DataCompoundStatement:removeChild]  Warning:  "
			   + "Unable to remove child (\""
			   + theChildToRemove . getClass() . getName()
			   + "\")." );
      return false;
    }
  }


  public void runOnSubcomponentFraction (
			String                      theString,
			RunOnSubcomponentInterface  theRunOnSubcomponentObject,
			Object                      theArgumentObject )
  {
	/* Don't search children by default. */
    runOnSubcomponentFraction ( theString,         theRunOnSubcomponentObject,
				theArgumentObject, false );
  }


  public DataComponent newComponent ( )
  {
    return new DataCompoundStatement ( null );
  }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );
    return theTDLParser . parseCompoundStatement ( 0 );
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    if ( /*NOT*/ !  ( theObjectToCopy instanceof DataCompoundStatement ) )
    {
      System.err.println ( "[DataCompoundStatement:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }


    return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
    String  subcomponent = null;


       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex();

	/* Generate any labels that we have... */
    generateLabels ( theOutputDestination, theObjectSubsetToGenerate,
		     DataCompoundStatement.OPEN_BRACE );

	/* Write any pre-open-brace non-significant tokens */
    generateSubcomponents ( DataCompoundStatement.OPEN_BRACE,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write our open-brace */
    theOutputDestination . write ( DataCompoundStatement.OPEN_BRACE );

	/* Increase our indenting... */
    theOutputDestination . addIndent ( DataComponent.STANDARD_INDENT );

	/* Write all the tokens up to the next-to-close-brace token... */
    generateSubcomponents ( getIndex ( DataCompoundStatement.CLOSE_BRACE ) - 1,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );


	/* At this point, we are going to do some fancy footwork.
	 * This is specifically to deal with the case where the close-brace
	 * is preceeded by a newline followed by spaces.  This causes
	 * the close-brace to be improperly indented.  The solution is to
	 * un-indent prior to writing this string, IF it exists...
	 */

	/* Is there another token before the CLOSE_BRACE,
	 * and it is a string consisting of just whitespace??
	 */
    if (    getHasSubcomponentsToGenerateBefore (
					  DataCompoundStatement.CLOSE_BRACE )
	 && isSubcomponentAString (
			 getIndex ( DataCompoundStatement.CLOSE_BRACE ) - 1 ) )
    {
      subcomponent = getStringSubcomponent (
			 getIndex ( DataCompoundStatement.CLOSE_BRACE ) - 1 );

      for ( int i=0; i < subcomponent.length();  i++ )
      {
	if ( Character.isWhitespace ( subcomponent . charAt ( i ) ) == false )
	{
	  subcomponent = null;
	  break;
	}
      }
    }


	/* If there was a string of whitespace before the close-brace */
    if ( subcomponent != null )
    {
	/* Write the line number of the close-brace */
      generateRelevantLineNumberMacros ( theOutputDestination,
					 theObjectSubsetToGenerate,
					 getLineNumberOfCloseBrace(),
					 DataCompoundStatement.CLOSE_BRACE );
        /* Un-indent now */
      theOutputDestination . removeIndent ( DataComponent.STANDARD_INDENT );
    }

	/* Write any remaining next-to-close-brace token... */
    generateSubcomponents ( DataCompoundStatement.CLOSE_BRACE,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );


	/* If there was NOT a string of whitespace before the close-brace, */
    if ( subcomponent == null )
    {
	/* Write the line number of the close-brace */
      generateRelevantLineNumberMacros ( theOutputDestination,
					 getLineNumberOfCloseBrace(),
					 0 );
	/* We should still un-indent. */
      theOutputDestination . removeIndent ( DataComponent.STANDARD_INDENT );
    }


	/* Write our close-brace */
    theOutputDestination . write ( DataCompoundStatement.CLOSE_BRACE );

	/* Write any remaining non-significant tokens */
	/* (There *SHOULD* *NOT* be any...  But just in case...) */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }



	/********************************************************/
        /********** Methods to interface with Graphing **********/
	/********************************************************/

  public boolean supportsGraphType ( int theGraphType )
  {
    if (   ( theGraphType == GraphRoot.GUI_TREE_GRAPH ) )
      return true;
    else
      return false;
  }

  public GraphModule createStatementGraphModule ( int theGraphType )
  {
    if (   ( theGraphType != GraphRoot.GUI_TREE_GRAPH ) )
      return null;

    /* else ... */

    /* Building:  +-{ }
     *            |  |
     */

    GraphConnection toChildren = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );

    GraphNode       statementNode = new GraphNode ( new SimpleLabel( "{...}" ),
				             GraphComponent.LEFT,
				DataCompoundStatement.COMPOUND_STATEMENT_COLOR,
					     false, toChildren );

    GraphConnection toStatement = new GraphConnection ( GraphComponent.LEFT,
				DataCompoundStatement.COMPOUND_STATEMENT_COLOR,
					     false, false, statementNode );

    GraphMeet       startMeet = new GraphMeet ( GraphComponent.UP,
				DataCompoundStatement.COMPOUND_STATEMENT_COLOR,
					    GraphComponent.LEFT, toStatement );

    GraphConnection tail = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );
    startMeet . addChild ( tail );

    GraphModule  graphModule = new GraphModule ( startMeet, tail, toChildren,
						 null, this, true,
						 theGraphType );

    updateGraphModule ( this, null, graphModule );

    return graphModule;
  }

	/* Invoked by this.update(*) invoking GraphModule.update() */
  public boolean updateGraphModule ( DataComponent  theModifiedComponent,
				     Object         theStoredData,
				     GraphModule    theGraphModule )
  {
	/* Skip updates to parents... */
    if ( theModifiedComponent != this )
      return true;

    return updateGraphModule ( theGraphModule, GraphModule.CHILD_COMPONENT );
  }



  	/** CompoundStatements are frequently treated as subparts of other
	  * classes.  And while they can have their own graph-modules,
	  * they usually serve as organizers of lists in other graph-modules.
	  *
	  * Thus, DataCompoundStatement's have a second updateGraphModule()
	  * method that does most of the real work in synchronizing
	  * DataComponent's with their corresponding GraphModules...
	  */
  public boolean updateGraphModule ( GraphModule  theGraphModule,
				     int          theChildIndex  )
  {
    GraphConnection    theGraphModuleChildConnection;
    GraphConnection    tailConnection  = null;
    GraphModule        graphModule     = theGraphModule,
		       oldGraphModule  = null,
                       nextGraphModule;
    GraphComponent     graphComponent;
    boolean            didDisableUpdates;
    int                i;

    
	/* Erase the modulesVector */
    getModulesVector() . removeAllElements();


	/* Find our First Subcomponent-Graph-Module */

	/* Verify that we have a GraphConnection tail to work with */
    if ( ( theGraphModule . getChild (theChildIndex) instanceof GraphConnection
	  ) == false )
    {
      System.err.println (
	    "[DataCompoundStatement:updateGraphModule]  Error:  "
	    + "theGraphModule.getChildComponent() is not a GraphConnection.   "
	    + "Aborting..." );
      return false;
    }

	/* Get the location of this DataCompoundStatement on the Graph */
    theGraphModuleChildConnection
      = (GraphConnection) (theGraphModule . getChild ( theChildIndex ));


      /* Freeze the Graph Hiearchy for modifications */
    didDisableUpdates = theGraphModule.disableUpdates ( false );


	/* Start us off at theGraphModuleChildConnection */
    tailConnection = theGraphModuleChildConnection;

	/* Iterate through all of our SubComponent's */
    for ( i = getIndex ( DataCompoundStatement.OPEN_BRACE  );
	  i < getIndex ( DataCompoundStatement.CLOSE_BRACE );
	  i++ )
    {
	/* Skip over non-DataComponents...*/
      if ( isSubcomponentADataComponent ( i ) == false )
	continue;

	/* Get the start of the next module */
      graphComponent  = tailConnection . getDestination();

	/* Get the next GraphModule */
      if ( graphComponent != null )
      {
	oldGraphModule = graphModule;
	graphModule    = graphComponent . getOwningGraphModule();

	  /* Do we have a problem? */
	if ( graphModule == oldGraphModule )
	{
	  System.err.println (
	    "[DataCompoundStatement:updateGraphModule]  Error:  "
	    + "Internal Inconsistency: GraphComponent (" + graphComponent
	    + ") after GraphModule (" + oldGraphModule
	    + ") is part of that GraphModule, not a new GraphModule..." );
	  return false;
	}

	  /* Do we have a big problem? */
	if ( graphModule == null )
	{
	  System.err.println (
	    "[DataCompoundStatement:updateGraphModule]  Warning:  "
	    + "Internal Inconsistency: GraphComponent (" + graphComponent
	    + ") after GraphModule (" + oldGraphModule
	    + ") is not part of any GraphModule..." );
	}
      } /* if ( graphComponent != null ) */


	/* If the DataComponents don't match... */
      if (   ( graphComponent != null )
	  && ( graphModule    != null )
	  && (    graphModule . getDataComponent()
	       != getDataComponentSubcomponent ( i ) ) )
      {
	   /* Does this GraphModule exist further on down? */
	nextGraphModule
	  = graphModule . findGraphModule ( getDataComponentSubcomponent ( i ),
					    GraphModule.LAST_COMPONENT );

	   /* If we found a matching graphModule... */
	if ( nextGraphModule != null )
	{
		/* Remove the current (& all trailing) GraphModules
		 * (We will add the ones we need back in later on...
		 *  An approach that is vastly more efficient than
		 *  continuously moving everything around...)
		 */
	  graphModule . removeModuleSubtreeFromGraph ( getModulesVector() );
	}

	  /* One way or another, insert a GraphModule below. */
	graphModule = null;

      } /* DataComponents do not match... */


	/* If we don't have anything... */
      if (   ( graphComponent == null )
	  || ( graphModule    == null ) )
      { 
	    /* Try to find a pre-existing corresponding GraphModule */
	graphModule = getModuleFor ( getDataComponentSubcomponent ( i ) );

	    /* If there is no pre-existing corresponding GraphModule */
	if ( graphModule == null )
	{
	    /* Create a new GraphModule for this DataComponent */
	  graphModule
	    = getDataComponentSubcomponent ( i )
	       . createGraphModule ( theGraphModule . getGraphType() );
	}

	  /* Insert the GraphModule */
	graphModule . insertModuleAt ( tailConnection );
      }


      /* Else: they match up, so just go on to the next component */


	/* Verify that we have a GraphConnection tail to work with */
      if ( graphModule . getLastComponent() == null )
      {
	System.err.println (
	    "[DataCompoundStatement:updateGraphModule]  Error:  "
	    + "graphModule.getLastComponent() is NULL.  Aborting..." );

		/* Restore the Graph Hiearchy and have it update */
	if ( didDisableUpdates )
	  theGraphModule.reenableUpdates ( true );

	return false;
      }

      else if ( ((graphModule . getLastComponent()) instanceof GraphConnection)
		== false )
      {
	System.err.println (
	    "[DataCompoundStatement:updateGraphModule]  Error:  "
	    + "graphModule.getLastComponent() is not a GraphConnection.  (\""
	    +  graphModule . getLastComponent() . getClass() . getName()
	    + "\")  Aborting..." );

		/* Restore the Graph Hiearchy and have it update */
	if ( didDisableUpdates )
	  theGraphModule.reenableUpdates ( true );

	return false;
      }

      else
      {
	  /* Get the tail */
	tailConnection = (GraphConnection)( graphModule . getLastComponent() );
      }

    } /* FOR ( all contained DataComponents ) */



	/* While there are still modules left, delete them... */

	/* Get the start of the next module */
    graphComponent  = tailConnection . getDestination();

	/* Get the next GraphModule */
    if ( graphComponent != null )
      nextGraphModule = graphComponent . getOwningGraphModule();
    else
      nextGraphModule = null;

    if (   ( nextGraphModule != null        )
	&& ( nextGraphModule != graphModule ) )
    {
	/* Remove the current (& all trailing) GraphModules
	 * in the most efficient manner possible...  (Ie: Bottom-First)
	 */
      getModulesVector() . removeAllElements();
      nextGraphModule . removeModuleSubtreeFromGraph ( getModulesVector() );
      getModulesVector() . removeAllElements();
    }

    else if ( graphComponent != null )
    {
      System.err.println (
	    "[DataCompoundStatement:updateGraphModule]  WARNING:  "
	    + "Ends in GraphComponent that not part of a new GraphModule." );
    }


	/* Restore the Graph Hiearchy and have it update */
    if ( didDisableUpdates )
      theGraphModule.reenableUpdates ( true );

    return true;
  }

}
