
import netscape.application.*;
import netscape.util.*;
import netscape.constructor.*;
import java.io.File;
import java.io.IOException;


public class GenerateHTMLWindowHandler extends FileTaskWindowHandler
{
	/* Constant (Special) Filenames */
  public final static String GENERATE_HTML_WINDOW_PLAN_NAME
						  = "GenerateHTMLWindow.planb";

	/* TASK_WINDOW_PLAN Objects */
  public final static String CANCEL_BUTTON         = "CancelButton";
  public final static String GENERATE_BUTTON       = "GenerateButton";
  public final static String NEW_FILE_BUTTON       = "NewFileButton";
  public final static String HTML_OUTPUT_TEXTFIELD = "HTMLOutputTextField";

	/* Commands */
  public final static String CANCEL_COMMAND        = "Cancel";
  public final static String GENERATE_COMMAND      = "GenerateHTML";
  public final static String NEW_FILE_COMMAND      = "NewFile";

	/* Constants */
  public final static String TITLE = "Generate HTML Task Documentation";


	/* Class variables */
  protected static Plan                       generateHTMLWindowPlan    = null;
  protected static GenerateHTMLWindowHandler  generateHTMLWindowHandler = null;


	/* Class Methods */
  protected static synchronized Plan getGenerateHTMLWindowPlan()
  {
    if ( GenerateHTMLWindowHandler.generateHTMLWindowPlan == null )
    {
      GenerateHTMLWindowHandler.generateHTMLWindowPlan
	= WindowHandlerAbstractBaseClass.getPlan (
		GenerateHTMLWindowHandler.GENERATE_HTML_WINDOW_PLAN_NAME,
		"[GenerateHTMLWindowHandler:getGenerateHTMLWindowPlan]  "
		+ "Error:  Failure - Unable to access "
		+ "GENERATE_HTML_WINDOW_PLAN_NAME" );
    }
    return GenerateHTMLWindowHandler.generateHTMLWindowPlan;
  }

  public static GenerateHTMLWindowHandler getGenerateHTMLWindowHandler()
  {
    if ( GenerateHTMLWindowHandler.generateHTMLWindowHandler == null )
    {
      GenerateHTMLWindowHandler.generateHTMLWindowHandler
	= new GenerateHTMLWindowHandler();
    }
    return GenerateHTMLWindowHandler.generateHTMLWindowHandler;
  }




	/* Instance Variables */
  protected Button               newFileButton,
                                 cancelButton,
		                 generateButton;
  protected TextField            outputFileTextfield;
  protected DataDestinationFile  dataDestinationFile;


	/* Constructors */
  public GenerateHTMLWindowHandler ( )
			{ createAndInitialize ( false ); }

  public GenerateHTMLWindowHandler ( boolean  theUseExternalWindow )
			{ createAndInitialize ( theUseExternalWindow ); }

	/* Constructors can't be synchronized. */
	/* And this needs to be synchronized since we use the static plans */
  private synchronized
      boolean createAndInitialize ( boolean theUseExternalWindow )
  {
	/* Initialize dataDestinationFile */
    dataDestinationFile = new DataDestinationFile();


	/* Let the super-class do it's thing... */
    if ( super . createAndInitialize (
			 GenerateHTMLWindowHandler.getGenerateHTMLWindowPlan(),
			 GenerateHTMLWindowHandler.GENERATE_COMMAND,
			 null,
			 theUseExternalWindow,
			 false,
			 GenerateHTMLWindowHandler.TITLE,
			 "GenerateHTMLWindowHandler",
			 null )
	 == false )
    {
      return false;
    }

	/* Lets find our instance variables */
    newFileButton
      = (Button)
         GenerateHTMLWindowHandler.getGenerateHTMLWindowPlan().componentNamed (
			         GenerateHTMLWindowHandler . NEW_FILE_BUTTON );
    cancelButton
      = (Button)
         GenerateHTMLWindowHandler.getGenerateHTMLWindowPlan().componentNamed (
				   GenerateHTMLWindowHandler . CANCEL_BUTTON );
    generateButton
      = (Button)
         GenerateHTMLWindowHandler.getGenerateHTMLWindowPlan().componentNamed (
				 GenerateHTMLWindowHandler . GENERATE_BUTTON );
    outputFileTextfield
      = (TextField)
         GenerateHTMLWindowHandler.getGenerateHTMLWindowPlan().componentNamed (
			   GenerateHTMLWindowHandler . HTML_OUTPUT_TEXTFIELD );

	/* Error-check */
    if (   ( newFileButton       == null )
	|| ( cancelButton        == null )
	|| ( generateButton      == null )
	|| ( outputFileTextfield == null ) )
    {
      System.err.println ( "[GenerateHTMLWindowHandler:createAndInitialize]  "
	   + "Warning:  One or more Plan objects is NULL!!!\n"
	   + "    newFileButton..........: " + newFileButton       + "\n"
	   + "    cancelButton...........: " + cancelButton        + "\n"
	   + "    generateButton.........: " + generateButton      + "\n"
	   + "    outputFileTextfield....: " + outputFileTextfield + "\n"
	   + "------------------------------------------------------------" );
    }


	/* Reset the plan so that we can use it again later on... */
    GenerateHTMLWindowHandler.getGenerateHTMLWindowPlan() . releaseObjects();

	/* Reconfigure our objects... */
    DefaultTextFilter.setDefaultTextFilter ( outputFileTextfield );

	/* Report that we succeeded. */
    return true;
  }



  public Button        getNewFileButton()       { return newFileButton;       }
  public Button        getCancelButton()        { return cancelButton;        }
  public Button        getGenerateButton()      { return generateButton;      }
  public TextField     getOutputFileTextfield() { return outputFileTextfield; }
  public DataDestinationFile
		       getDataDestinationFile() { return dataDestinationFile; }


	/* ExtendedTarget Interface */
  public boolean canPerformCommand ( String theCommand ) 
  {
    return 
      (    theCommand . equals ( GenerateHTMLWindowHandler.CANCEL_COMMAND   )
	|| theCommand . equals ( GenerateHTMLWindowHandler.GENERATE_COMMAND )
	|| theCommand . equals ( GenerateHTMLWindowHandler.NEW_FILE_COMMAND )
	|| super . canPerformCommand ( theCommand ) );
  }


	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
	/* Handle canceling... */
    if ( theCommand . equals ( GenerateHTMLWindowHandler.CANCEL_COMMAND ) )
    {
      getWindow() . hide(); /* Close the window */
    }

	/* Handle successful conclusion. */
    else if ( theCommand
	        . equals ( GenerateHTMLWindowHandler.GENERATE_COMMAND ) )
    {
      if ( generateHTML() )
	getWindow() . hide();  /* Close the window... */
    }


	/* Handle a new-file choice */
    else if ( theCommand
	        . equals ( GenerateHTMLWindowHandler.NEW_FILE_COMMAND ) )
    {
      DataFile     dataFile;
      FileChooser  fileChooser = new FileChooser ( getRootView(),
						   "Select File",
						   FileChooser.LOAD_TYPE );

      if ( WindowHandlerAbstractBaseClass.getPriorSelectedDirectory() != null )
      {
	fileChooser . setDirectory (
		  WindowHandlerAbstractBaseClass.getPriorSelectedDirectory() );
      }

      fileChooser . showModally();

      WindowHandlerAbstractBaseClass.setPriorSelectedDirectory (
						   fileChooser . directory() );

	/* Did the user "CANCEL" the fileChooser? */
      if ( fileChooser . file() == null )
	return;

      dataFile = WindowHandlerAbstractBaseClass.validateNewFilename (
				fileChooser.directory() + fileChooser.file(),
				getRootView(), false );

	/* Are we done?  Did validateNewFilename terminate unsuccessfully? */
      if ( dataFile == null )
	return;

      loadKnownFiles();
      selectFile ( dataFile );
    }

	/* Otherwise, let the superclass deal with it... */
    else
    {
      super . performCommand ( theCommand, theData );
    }

  } /* public void performCommand ( String theCommand, Object theData ) */


  public boolean generateHTML ()
  {
    String       outputFilename = getOutputFileTextfield() . stringValue();
    DataFile     selectedDataFile;
    File         existFile;

	/* Do we have a selected file? */
    if (   ( getFileListView() . selectedItem() == null )
	|| ( getFileListView() . count()        <= 0    ) )
    {
      runAlert ( "Error:", "No File has been Selected.",
		 "OK", null, null );
      return false;
    }

	/* Do we have someplace to generate to? */
    if ( outputFilename . length() <= 0 )
    {
      runAlert ( "Error:", "No HTML Output file has been specified.",
		 "Abort", null, null );
      return false;
    }

	/* Find our corresponding selected DataFile */
    selectedDataFile = (DataFile)(getFileListView() . selectedItem() . data());

	/* And verify it... */
    if ( selectedDataFile == null )
    {
      System.err.println ( "[GenerateHTMLWindowHandler:generateHTML]  "
			   + "Error:  selectedDataFile == null." );
      return false;
    }

	/* Deal with Absolute/Relative path names... */
    outputFilename
      = WindowHandlerAbstractBaseClass.resolveFilepath (
					     outputFilename,
					     selectedDataFile . getFilename(),
					     "output file.  Generate HTML to",
					     getRootView() );

	/* Did the user ABORT in resolveFilepath()? */
    if ( outputFilename == null )
      return false;


	/* Does the output file already exist? */
    existFile = new File ( outputFilename );
    if ( existFile . isDirectory() )
    {
      runAlert ( "Error:",
		 "Output file \"" + outputFilename + "\" is a directory!!!",
		 "Abort", null, null );
      return false;
    }
    else if ( existFile . exists() )
    {
      if ( runAlert ( "Warning:",
		      "Output file \"" + outputFilename
		      + "\" already exists!!!\n\nDelete \""
		      + outputFilename + "\"?",
		      "Cancel", "Delete", null )
	   != Alert.SECOND_OPTION )
      {
	return false;
      }
      else
      {
	if ( existFile . delete() == false )
	  System.err.println ( "[GenerateHTMLWindowHandler:generateHTML]  "
			       + "Warning:  Deletion of \"" + outputFilename
			       + "\" failed." );
      }
    }


	/* Try to open the file... */
    try
    {
      getDataDestinationFile()
	. setFileWithoutExceptionHandling ( outputFilename );
    }
    catch ( IOException  theException )
    {
      runAlert ( "Error:",
		 "Unable to open file \"" + outputFilename
		 + "\".\n\n  (IOException = " + theException + ")",
		 "Abort", null, null );
      return false;
    }
    catch ( SecurityException  theException )
    {
      runAlert ( "Error:",
		 "Unable to open file \"" + outputFilename
		 + "\".\n\n  (SecurityException = " + theException + ")",
		 "Abort", null, null );
      return false;
    }


	/* Try to write the file... */
    selectedDataFile . generate ( getDataDestinationFile(),
				  DataComponent.HTML_DOCUMENTATION );


	/* Try to close the file... */
    try
    {
      getDataDestinationFile() . closeWithoutExceptionHandling();
    }
    catch ( IOException  theException )
    {
      runAlert ( "Error:",
		 "Unable to close file \"" + outputFilename
		 + "\".\n\n  (IOException = " + theException + ")",
		 "Abort", null, null );
      return false;
    }

    return true;
  }



  public void performSetFile ( )
  {
    DataFile     selectedDataFile;

	/* Do we have a selected file? */
    if (   ( getFileListView() . selectedItem() != null )
	&& ( getFileListView() . count()         > 0    ) )
    {
	/* Find our corresponding selected DataFile */
      selectedDataFile
	= (DataFile) (getFileListView() . selectedItem() . data());

	/* If we have a base-filename to work off of... */
      if (   ( selectedDataFile != null )
	  && ( DataComponent.isEmptyString ( selectedDataFile . getFilename() )
	       == false ) )
      {
	getOutputFileTextfield() . setStringValue (
		   DataFile.translateTDLFilenameToCxx (
					  selectedDataFile . getFilename(),
					  DataFile.STANDARD_HTML_TAIL      ) );
      }
      else
      {
	getOutputFileTextfield() . setStringValue ( "" );
      }
    }
    else
    {
      getOutputFileTextfield() . setStringValue ( "" );
    }

    super . performSetFile();
  }

}
