// *DBA-Alert*:  Need to set DataFile when editing Tasks...


import netscape.application.*;
import netscape.util.*;
import netscape.constructor.*;

public class NewTaskWindowHandler extends WindowHandlerAbstractBaseClass
			       implements DataComponentObserver,
					  ExtendedTarget, Target,
					  TextFieldOwner, WindowOwner
{
	/* Constant (Special) Filenames */
  public final static String TASK_WINDOW_PLAN_NAME   = "NewTaskWindow.planb";
  public final static String TASK_ARGUMENTS_PLAN_NAME= "NewTaskArgument.planb";

	/* TASK_WINDOW_PLAN Objects */
  public final static String TASK_BUTTON_NAME       = "TaskButton";
  public final static String GOAL_BUTTON_NAME       = "GoalButton";
  public final static String COMMAND_BUTTON_NAME    = "CommandButton";
  public final static String EXCEPTION_BUTTON_NAME  = "ExceptionButton";
  public final static String MONITOR_BUTTON_NAME    = "MonitorButton";
  public final static String EXTERN_CHECKBOX_NAME   = "ExternCheckbox";
  public final static String PLAN_FIRST_CHECKBOX_NAME
						    = "PlanFirstCheckbox";
  public final static String DELAY_PLANNING_CHECKBOX_NAME
						    = "DelayPlanningCheckbox";
  public final static String SELECT_FILE_NAME       = "SelectFileButton";
  public final static String FILE_LOCATION_NAME     = "FileLocationTextField";
  public final static String TASK_TEXTFIELD_NAME    = "TaskName";
  public final static String ARGUMENTS_SCROLL_GROUP_NAME
						    = "ArgumentsScrollGroup";
  public final static String ADD_MORE_ARGUMENTS_BUTTON_NAME
						    = "AddMoreArgumentsButton";
  public final static String SCROLL_TEXT_NAME       = "ScrollTextData";
  public final static String INTERNAL_WINDOW_BUTTON = "InternalWindowButton";
  public final static String EXTERNAL_WINDOW_BUTTON = "ExternalWindowButton";
  public final static String NO_WINDOW_BUTTON       = "NoWindowButton";
  public final static String CANCEL_BUTTON          = "CancelButton";
  public final static String OK_BUTTON              = "OKButton";


	/* Arguments Plan Objects */
  public final static String ARGUMENT_GROUP_NAME     = "ArgumentGroup";
  public final static String IN_OUT_BUTTON_NAME      = "InOutButton";
  public final static String ARGUMENT_TEXTFIELD_NAME = "ArgumentTextField";

	/* Commands */
  public final static String OK_COMMAND                  = "OK";
  public final static String CANCEL_COMMAND              = "Cancel";
  public final static String ADD_ARGUMENTS_COMMAND       = "AddArguments";
  public final static String DATA_CHANGED_COMMAND        = "DataChanged";
  public final static String SELECT_FILE_COMMAND         = "SelectFile";
  public final static String TOGGLE_IN_OUT_BUTTON_COMMAND= "ToggleInOutButton";

	/* Constants */
  public final static String[] INOUT_BUTTON_TITLES  /* IN, OUT, INOUT */
					     = DataTaskArgument.ARGUMENT_TYPES;
  public final static String   NEW_TASK_TITLE      = "Create New Task";
  public final static String   EXISTING_TASK_TITLE = "Edit Task Definition";

  public final static int      ARGUMENTS_PADDING     = 5;
  public final static Font     CODE_SCROLL_VIEW_FONT
				= Font.fontNamed ( "Courier", Font.PLAIN, 10 );




	/* Class variables */
  protected static Plan newTaskWindowPlan    = null,
		        newTaskArgumentsPlan = null;

	/* Class Methods */
  protected static synchronized Plan getNewTaskWindowPlan()
  {
    if ( NewTaskWindowHandler.newTaskWindowPlan == null )
    {
      NewTaskWindowHandler.newTaskWindowPlan
	= WindowHandlerAbstractBaseClass.getPlan (
		NewTaskWindowHandler.TASK_WINDOW_PLAN_NAME,
		"[NewTaskWindowHandler:getNewTaskWindowPlan]  "
		+ "Error:  Failure - Unable to access "
		+ "TASK_WINDOW_PLAN_NAME" );
    }
    return NewTaskWindowHandler.newTaskWindowPlan;
  }

  protected static synchronized Plan getNewTaskArgumentsPlan()
  {
    if ( NewTaskWindowHandler.newTaskArgumentsPlan == null )
    {
      NewTaskWindowHandler.newTaskArgumentsPlan
	= WindowHandlerAbstractBaseClass.getPlan (
		NewTaskWindowHandler.TASK_ARGUMENTS_PLAN_NAME,
		"[NewTaskWindowHandler:getNewTaskArgumentsPlan]  "
		+ "Error:  Failure - Unable to access "
		+ "TASK_ARGUMENTS_PLAN_NAME" );
    }
    return NewTaskWindowHandler.newTaskArgumentsPlan;
  }

  protected static boolean InitPlans ( )
  {
	/* Make certain our font-of-choice is loaded & available. */
    Font makeSureThisIsLoadedAndAvailable = CODE_SCROLL_VIEW_FONT;

	/* Try to load our plans... */
    boolean  foundNewTaskWindowPlan    = ( getNewTaskWindowPlan()    != null ),
             foundNewTaskArgumentsPlan = ( getNewTaskArgumentsPlan() != null );

    return foundNewTaskWindowPlan && foundNewTaskArgumentsPlan;
  }




	/* Instance Variables */
  protected Button                     taskButton, goalButton, commandButton,
			               exceptionButton, monitorButton,
			               externCheckbox, planFirstCheckbox,
			               delayPlanningCheckbox,
				       selectFileButton;
  protected TextField                  locationFileTextField;
  protected TextField                  taskNameTextField;
  protected ScrollGroup                argumentsScrollGroup;
  protected Button                     addMoreArgumentsButton;
  protected ScrollTextOfDataComponent  plainCodeScrollText;
  protected Button                     internalWindowButton,
				       externalWindowButton,
				       noWindowButton,
				       cancelButton, okButton;
  protected ContainerView              argumentsContainer;

  protected Vector                     argumentSubviews;
  protected DataTaskDefinition         dataTaskDefinition;
  protected DataFile                   locationDataFile;


	/* Constructors */
  public NewTaskWindowHandler ( ) { createAndInitialize ( false, null ); }

  public NewTaskWindowHandler ( boolean  theUseExternalWindow )
		      { createAndInitialize ( theUseExternalWindow, null ); }

  public NewTaskWindowHandler ( DataTaskDefinition  theDataTaskDefinition )
		      { createAndInitialize ( false, theDataTaskDefinition ); }

  public NewTaskWindowHandler ( boolean             theUseExternalWindow,
				DataTaskDefinition  theDataTaskDefinition )
       { createAndInitialize ( theUseExternalWindow, theDataTaskDefinition ); }


	/* Constructors can't be synchronized. */
	/* And this needs to be synchronized since we use the static plans */
  private synchronized
     boolean createAndInitialize ( boolean             theUseExternalWindow,
				   DataTaskDefinition  theDataTaskDefinition )
  {
	/* Do we have a Plan object? */
    if ( InitPlans() == false )
      return false;

	/* Let the superclass do the work... */
    if ( super . createAndInitialize (
			   NewTaskWindowHandler.getNewTaskWindowPlan(),
			   theUseExternalWindow,
			   (   ( theDataTaskDefinition == null )
			     ? NewTaskWindowHandler.NEW_TASK_TITLE
			     : NewTaskWindowHandler.EXISTING_TASK_TITLE ),
			   "NewTaskWindowHandler",
			   null )
	 == false )
    {
      return false;
    }

	/* Create our argumentSubviews Vector */
    argumentSubviews = new Vector ( );

	/* And initialize locationDataFile */
    locationDataFile = null;

	/* Lets find our instance variables */
    taskButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				     NewTaskWindowHandler . TASK_BUTTON_NAME );
    goalButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				     NewTaskWindowHandler . GOAL_BUTTON_NAME );
    commandButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				  NewTaskWindowHandler . COMMAND_BUTTON_NAME );
    exceptionButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				NewTaskWindowHandler . EXCEPTION_BUTTON_NAME );
    monitorButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				  NewTaskWindowHandler . MONITOR_BUTTON_NAME );
    externCheckbox
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				 NewTaskWindowHandler . EXTERN_CHECKBOX_NAME );
    planFirstCheckbox
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
			     NewTaskWindowHandler . PLAN_FIRST_CHECKBOX_NAME );
    delayPlanningCheckbox
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
			 NewTaskWindowHandler . DELAY_PLANNING_CHECKBOX_NAME );
    selectFileButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				     NewTaskWindowHandler . SELECT_FILE_NAME );
    locationFileTextField
      = (TextField)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				   NewTaskWindowHandler . FILE_LOCATION_NAME );
    taskNameTextField
      = (TextField)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				  NewTaskWindowHandler . TASK_TEXTFIELD_NAME );
    argumentsScrollGroup
      = (ScrollGroup)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
			  NewTaskWindowHandler . ARGUMENTS_SCROLL_GROUP_NAME );
    addMoreArgumentsButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
		       NewTaskWindowHandler . ADD_MORE_ARGUMENTS_BUTTON_NAME );
    plainCodeScrollText
      = (ScrollTextOfDataComponent)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				     NewTaskWindowHandler . SCROLL_TEXT_NAME );
    internalWindowButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
			       NewTaskWindowHandler . INTERNAL_WINDOW_BUTTON );
    externalWindowButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
			       NewTaskWindowHandler . EXTERNAL_WINDOW_BUTTON );
    noWindowButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
				     NewTaskWindowHandler . NO_WINDOW_BUTTON );
    cancelButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
					NewTaskWindowHandler . CANCEL_BUTTON );
    okButton
      = (Button)
         NewTaskWindowHandler.getNewTaskWindowPlan().componentNamed (
					    NewTaskWindowHandler . OK_BUTTON );



	/* Error-check */
    if (   ( taskButton             == null )
	|| ( goalButton             == null )
	|| ( commandButton          == null )
	|| ( exceptionButton        == null )
	|| ( monitorButton          == null )
	|| ( externCheckbox         == null )
	|| ( planFirstCheckbox      == null )
	|| ( delayPlanningCheckbox  == null )
	|| ( selectFileButton       == null )
	|| ( locationFileTextField  == null )
	|| ( taskNameTextField      == null )
	|| ( argumentsScrollGroup   == null )
	|| ( addMoreArgumentsButton == null )
	|| ( plainCodeScrollText    == null )
	|| ( internalWindowButton   == null )
	|| ( externalWindowButton   == null )
	|| ( noWindowButton         == null )
	|| ( cancelButton           == null )
	|| ( okButton               == null ) )
    {
      System.err.println ( "[NewTaskWindowHandler:createAndInitialize]  "
	   + "Warning:  One or more Plan objects is NULL!!!\n"
	   + "    taskButton.............: " + taskButton             + "\n"
	   + "    goalButton.............: " + goalButton             + "\n"
	   + "    commandButton..........: " + commandButton          + "\n"
	   + "    exceptionButton........: " + exceptionButton        + "\n"
	   + "    monitorButton..........: " + monitorButton          + "\n"
	   + "    externCheckbox.........: " + externCheckbox         + "\n"
	   + "    planFirstCheckbox......: " + planFirstCheckbox      + "\n"
	   + "    delayPlanningCheckbox..: " + delayPlanningCheckbox  + "\n"
	   + "    selectFileButton.......: " + selectFileButton       + "\n"
	   + "    locationFileTextField..: " + locationFileTextField  + "\n"
	   + "    taskNameTextField......: " + taskNameTextField      + "\n"
	   + "    argumentsScrollGroup...: " + argumentsScrollGroup   + "\n"
	   + "    addMoreArgumentsButton.: " + addMoreArgumentsButton + "\n"
	   + "    plainCodeScrollText....: " + plainCodeScrollText    + "\n"
	   + "    internalWindowButton...: " + internalWindowButton   + "\n"
	   + "    externalWindowButton...: " + externalWindowButton   + "\n"
	   + "    noWindowButton.........: " + noWindowButton         + "\n"
	   + "    cancelButton...........: " + cancelButton           + "\n"
	   + "    okButton...............: " + okButton               + "\n"
	   + "------------------------------------------------------------" );
    }



	/* Reset the plan so that we can use it again later on... */
    NewTaskWindowHandler.getNewTaskWindowPlan() . releaseObjects();


	/* Finish configuring our window */
    getWindow() . setOwner ( this );


	/* Perform object reconfigurations... */
    DefaultTextFilter.setDefaultTextFilter ( locationFileTextField );
    DefaultTextFilter.setDefaultTextFilter ( taskNameTextField );
    locationFileTextField . setOwner ( this );
    taskNameTextField     . setOwner ( this );
    plainCodeScrollText . setIsAutoResizing ( false );
    plainCodeScrollText . setFont (NewTaskWindowHandler.CODE_SCROLL_VIEW_FONT);
    addMoreArgumentsButton . removeFromSuperview();
    argumentsContainer = new ContainerView ( );
    argumentsContainer . setBorder ( null );
    argumentsContainer . setBackgroundColor ( Color.lightGray );
    argumentsContainer . setHorizResizeInstruction ( View.WIDTH_CAN_CHANGE );
    argumentsScrollGroup . setContentView ( argumentsContainer );
    argumentsContainer . addSubview ( addMoreArgumentsButton );
    addMoreArgumentsButton . moveTo ( NewTaskWindowHandler.ARGUMENTS_PADDING,
				      NewTaskWindowHandler.ARGUMENTS_PADDING );
    argumentsContainer . setBounds (
			  0, 0,
			  argumentsScrollGroup.scrollView() . bounds() . width,
			  addMoreArgumentsButton . bounds() . height
			  + (2 * NewTaskWindowHandler.ARGUMENTS_PADDING) );


	/* If we don't already have a set of task-data... */
    if ( theDataTaskDefinition == null )
    {
	/* Create a new Task-Data object. */
      setDataTaskDefinition ( new DataTaskDefinition ( ) );
      getDataTaskDefinition() . addSubcomponent ( " " );
      getDataTaskDefinition() . setIndex( DataTaskDefinition.TASK_TYPE_INDEX );
      getDataTaskDefinition() . addSubcomponent ( " " );
      getDataTaskDefinition() . setIndex( DataTaskDefinition.TASK_NAME_INDEX );
      getDataTaskDefinition() . addSubcomponent ( " " );      
      getDataTaskDefinition() . setIndex( DataTaskDefinition.OPEN_PAREN );
      getDataTaskDefinition() . addSubcomponent ( " " );      
      getDataTaskDefinition() . setIndex( DataTaskDefinition.CLOSE_PAREN );
    }
    else
    {
	/* Use the existing Task-Data object */
      setDataTaskDefinition ( theDataTaskDefinition );

	/* Stop showing window choices */
      noWindowButton . setState ( true );
      noWindowButton . superview() . removeFromSuperview();

	/* Remove the Cancel button */
      cancelButton . removeFromSuperview();

	/* And raise/center the OK button */
      okButton . moveTo ( noWindowButton . superview() . bounds() . x
			  + ( ( noWindowButton . superview() . bounds() . width
				- okButton . bounds() . width )
			      / 2 ),
			  noWindowButton . superview() . bounds() . y );

	/* Shorten the window... */
      getWindow()
	. sizeBy ( 0, - noWindowButton . superview() . bounds() . height );

	/* Make the file-location non-editable [for now *DBA-ALERT*]... */
      locationFileTextField . setEditable ( false );
      selectFileButton      . setEnabled  ( false );
    }


	/* Report that we succeeded. */
    return true;
  }


  public RootView    getRootView()
			        { return getTaskNameTextField() . rootView(); }

  public Button      getTaskButton()          { return taskButton; }
  public Button      getGoalButton()          { return goalButton; }
  public Button      getCommandButton()       { return commandButton; }
  public Button      getExceptionButton()     { return exceptionButton; }
  public Button      getMonitorButton()       { return monitorButton; }
  public Button      getExternCheckbox()      { return externCheckbox; }
  public Button      getPlanFirstCheckbox()   { return planFirstCheckbox; }
  public Button      getDelayPlanningCheckbox(){return delayPlanningCheckbox; }
  public TextField   getLocationFileTextField(){return locationFileTextField; }
  public TextField   getTaskNameTextField()   { return taskNameTextField; }
  public ScrollGroup getArgumentsScrollGroup(){ return argumentsScrollGroup; }
  public Button    getAddMoreArgumentsButton(){ return addMoreArgumentsButton;}
  public ScrollTextOfDataComponent
		     getPlainCodeScrollText() { return plainCodeScrollText; }
  public Button      getInternalWindowButton(){ return internalWindowButton; }
  public Button      getExternalWindowButton(){ return externalWindowButton; }
  public Button      getNoWindowButton()      { return noWindowButton; }
  public Button      getCancelButton()        { return cancelButton; }
  public Button      getOkButton()            { return okButton; }

  public DataFile    getLocationDataFile()    { return locationDataFile; }

  public void        setLocationDataFile ( DataFile theLocationDataFile )
  {
    if (   ( theLocationDataFile   != null                  )
	&& ( getLocationDataFile() != null                  )
	&& ( theLocationDataFile   != getLocationDataFile() )
	&& ( getLocationDataFile() . getIsEmpty()           ) )
    {
      getLocationDataFile() . unregisterFile (
				  "NewTaskWindowHandler:setLocationDataFile" );
    }

    locationDataFile = theLocationDataFile;

    if (   ( theLocationDataFile        != null )
	&& ( getLocationFileTextField() != null ) )
    {
      setTextFieldText ( getLocationFileTextField(),
			 theLocationDataFile . getFilename() );
    }
  }

  public ContainerView getArgumentsContainer(){ return argumentsContainer; }

  public Vector      getArgumentSubviews ( )  { return argumentSubviews; }

  public View        getArgumentSubview ( int theIndex )
  {
    if ( ( theIndex >= 0 )  && ( theIndex < getArgumentSubviews().count() ) )
      return (View) getArgumentSubviews() . elementAt ( theIndex );
    else
    {
      System.err.println ( "[NewTaskWindowHandler:getArgumentSubview]  "
			   + "Warning:  Invalid Index (" + theIndex
			   + "," + getArgumentSubviews().count() + ")" );
      return null;
    }
  }


  public DataTaskDefinition getDataTaskDefinition ( )
						 { return dataTaskDefinition; }

  public void setDataTaskDefinition( DataTaskDefinition theDataTaskDefinition )
  {
	/* If we have a current dataTask object, clear us out of it. */
    if ( getDataTaskDefinition() != null )
      clearDataTaskDefinition();

	/* Set our new dataTask object. */
    dataTaskDefinition = theDataTaskDefinition;
    getDataTaskDefinition() . addObserver ( this, null );
    getPlainCodeScrollText()
      . setDataComponent ( getDataTaskDefinition(),
			   DataTaskDefinition.TASK_WITHOUT_BODY );

	/* Synchronize us with this data-task object. */
    getDataTaskDefinition() . update();
  }


  public boolean unregisterToObserve ()
  {
    return clearDataTaskDefinition();
  }


  public boolean clearDataTaskDefinition ( )
  {
    boolean  returnValue = true;

	/* Remove any children-observations... */
    for ( int i=0;  i < getDataTaskDefinition().getTaskArgumentCount();  i++ )
      if ( getDataTaskDefinition() . getTaskArgument( i )
	                           . removeObserver ( this ) == false )
	returnValue = false;

	/* And stop observing this task object. */
    if ( getDataTaskDefinition() . removeObserver ( this ) == false )
      returnValue = false;

    if ( getPlainCodeScrollText() . setDataComponent ( null ) == false )
      returnValue = false;

	/* And lets loose the dataTaskDefinition object... */
    dataTaskDefinition = null;

    return returnValue;
  }


	/** This routine takes advantage of the fact that we are storing the
	 *  argument subviews as the "data" component of
	 *  DataComponentObservable, thus giving us a mapping between
	 *  DataTaskArguments and their corresponding (observing) views.
	 *  (Which, by exhaustive search, can be reversed.)
	 */
  public DataTaskArgument findDataTaskArgumentForView ( View theSubview )
  {
    for ( int i=0;  i < getDataTaskDefinition().getTaskArgumentCount();  i++ )
    {
      if ( getDataTaskDefinition() . getTaskArgument( i )
				   . getObservableData ( this )
				   . getData()
	   == theSubview )
      {
	return getDataTaskDefinition() . getTaskArgument( i );
      }
    }
    return null;
  }



  public synchronized
	   View  addAdditionalArgument ( DataTaskArgument theDataTaskArgument )
  {
    NewTaskWindowArgumentView  newArgumentsView;
    Button                     inOutButton;
    TextField                  argument;
    int                        deltaWidth;

	/* Do we have a Plan object? */
    if ( NewTaskWindowHandler.getNewTaskArgumentsPlan() == null )
    {
      System.err.println ( "[NewTaskWindowHandler:addAdditionalArgument]  "
			   + "Error:  getNewTaskArgumentsPlan() == null.  "
			   + "  -- No Plan!" );
      return null;
    }

	/* Create a new set of argument objects... */
    if ( NewTaskWindowHandler.getNewTaskArgumentsPlan()
	   . unarchiveObjects ( this )
	 == false )
    {
      System.err.println ( "[NewTaskWindowHandler:addAdditionalArgument]  "
			   + "Warning:  "
			   + "newTaskArgumentsPlan.unarchiveObjects() "
			   + "returned false.   Probable failure." );
    }

	/* Find our new argument-set view. */
    newArgumentsView
      = (NewTaskWindowArgumentView)
         NewTaskWindowHandler.getNewTaskArgumentsPlan().componentNamed (
				  NewTaskWindowHandler . ARGUMENT_GROUP_NAME );
    inOutButton
      = (Button)
         NewTaskWindowHandler.getNewTaskArgumentsPlan().componentNamed (
				  NewTaskWindowHandler . IN_OUT_BUTTON_NAME );
    argument
      = (TextField)
         NewTaskWindowHandler.getNewTaskArgumentsPlan().componentNamed (
			      NewTaskWindowHandler . ARGUMENT_TEXTFIELD_NAME );

	/* Error check */
    if (   ( newArgumentsView == null )
	|| ( inOutButton      == null )
	|| ( argument         == null ) )
    {
      System.err.println ( "[NewTaskWindowHandler:addAdditionalArgument]  "
        + "Warning:  One or more elements of the new argument set is NULL!!!\n"
	+ "    newArgumentsView....: " + newArgumentsView + "\n"
	+ "    inOutButton.........: " + inOutButton      + "\n"
	+ "    argument............: " + argument         + "\n"
	+ "--------------------------------------------------------------" );
    }

	/* Store references to these objects for later use... */
    newArgumentsView . setInOutButton ( inOutButton );
    newArgumentsView . setArgument    ( argument    );


	/* Establish "nice" text-field editing... */
    DefaultTextFilter.setDefaultTextFilter ( argument );

	/* And make certain that we know when this TextField changes... */
    argument . setOwner  ( this );


	/* Configure objects. */
    newArgumentsView . setHorizResizeInstruction ( View.WIDTH_CAN_CHANGE );


	/* Position everything */
    newArgumentsView . moveTo (
			    NewTaskWindowHandler.ARGUMENTS_PADDING,
			    (   getArgumentsContainer()     . bounds() . height
			      - getAddMoreArgumentsButton() . bounds() . height
			      - NewTaskWindowHandler.ARGUMENTS_PADDING ) );
    getAddMoreArgumentsButton() . moveBy (
			    0,
			    (  newArgumentsView . bounds() . height
			     + NewTaskWindowHandler.ARGUMENTS_PADDING ) );

	/* Resize everything */
    deltaWidth =   newArgumentsView . bounds() . width
		 + (2 * NewTaskWindowHandler.ARGUMENTS_PADDING)
                 - getArgumentsContainer() . bounds() . width;
    if ( deltaWidth < 0 )
    {
      newArgumentsView . sizeBy ( - deltaWidth, 0 );
      deltaWidth = 0;
    }
    getArgumentsContainer() . sizeBy (
				deltaWidth,
				(   newArgumentsView . bounds() . height
				  + NewTaskWindowHandler.ARGUMENTS_PADDING ) );

	/* Put it in our view hierarchy */
    getArgumentsContainer() . addSubview ( newArgumentsView );
	/* And keep a record that we have this view */
    getArgumentSubviews() . addElement ( newArgumentsView );

	/* And redraw it. */
    getArgumentsContainer() . setDirty ( true );
    getArgumentsContainer()
      . scrollRectToVisible ( getAddMoreArgumentsButton() . bounds() );


	/* reset this plan so we can use it again later on... */	
    NewTaskWindowHandler.getNewTaskArgumentsPlan() . releaseObjects();


	/* Make certain that we are not about to suffer painfully */
    if ( getDataTaskDefinition() == null )
    {
      System.err.println ( "[NewTaskWindowHandler:addAdditionalArgument]  "
			   + "Error:  getDataTaskDefinition() is null.  "
			   + "(Creating new DataTaskDefinition object...)" );
      setDataTaskDefinition ( new DataTaskDefinition() );
    }

	/* Create a DataTaskArgument object if we need to... */
    if ( theDataTaskArgument == null )
    {
      theDataTaskArgument = new DataTaskArgument();
      theDataTaskArgument . addSubcomponent ( " " );
      theDataTaskArgument . setIndex (
				   DataTaskArgument.ARGUMENT_DIRECTION_INDEX );

	/* Assume this task argument is already in the task-definition */
	/* if it was passed in... */
      getDataTaskDefinition() . addTaskArgument ( theDataTaskArgument );
    }

	/* Have us be notified when this object changes. */
    theDataTaskArgument . addObserver ( this, newArgumentsView );
    
	/* And synchronize everyone with the DataComponents... */
    theDataTaskArgument . update ( null );

    return newArgumentsView;
  }


  public View removeLastArgumentSubview ( )
  {
    int   index           = getArgumentSubviews() . count()  -  1;
    View  argumentSubview;

	/* Do we have any argument-subviews to remove? */
    if ( index < 0 )
    {
      System.err.println ( "[NewTaskWindowHandler:removeLastArgumentSubview]  "
			   + "Warning:  No arguments to remove." );
      return null;
    }

	/* Find our argument subview */
    argumentSubview = getArgumentSubview ( index );

	/* Remove it from the view hierarchy */
    argumentSubview . removeFromSuperview();

	/* And from our internal reference list */
    getArgumentSubviews() . removeElementAt ( index );

	/* Adjust the other views correspondingly */
    addMoreArgumentsButton . moveBy ( 0,
				      - ( argumentSubview . bounds() . height )
				      - NewTaskWindowHandler.ARGUMENTS_PADDING
				     );
    argumentsContainer     . sizeBy ( 0,
				      - ( argumentSubview . bounds() . height )
				      - NewTaskWindowHandler.ARGUMENTS_PADDING
				     );
    argumentsContainer . setDirty ( true );

	/* And we are done */
    return argumentSubview;
  }





	/* DataComponentObserver Interface */
  public boolean update ( DataComponent  theObservedComponent,
			  DataComponent  theModifiedComponent,
			  Object         theStoredData        )
  {
    int  i;

 	/* Ignore any child modification messages that have been */
	/* propagated up to the parent observers... */
    if ( theObservedComponent != theModifiedComponent )
      return true;


	/* Was an argument modified? */
    if ( theObservedComponent instanceof DataTaskArgument )
    {
      if ( argumentsContainer . subviews() . containsIdentical( theStoredData )
	   == false )
      {
	System.err.println ( "[NewTaskWindowHandler:update]  Error:  "
			     + "DataTaskArgument does not refer to a known "
			     + "subview.  Failing..." );
	return false;
      }

      ( (NewTaskWindowArgumentView) theStoredData ) . getInOutButton()
	. setTitle ( ( (DataTaskArgument) theObservedComponent )
		     . getArgumentDirectionString() );

      ( (NewTaskWindowArgumentView) theStoredData ) . getArgument()
	. setStringValue ( ( (DataTaskArgument) theObservedComponent )
			   . toString ( DataTaskArgument.TYPE_NAME_AND_EQUALS )
			  );

	/* Note that we have succeeded */
      return true;
    }


	/* Was the DataTaskDefinition modified? */
    if ( theObservedComponent == getDataTaskDefinition() )
    {
	/* Declare some convenience variables to simplify things */
      DataTaskDefinition          dataTaskDef
				   = (DataTaskDefinition) theObservedComponent;
      DataComponentObservableData argumentObservableData;


      switch ( dataTaskDef . getTaskType() )
      {
	case DataTaskDefinition.TASK_TASK:
	  getTaskButton() . setState ( true );
	  break;
	case DataTaskDefinition.GOAL_TASK:
	  getGoalButton() . setState ( true );
	  break;
	case DataTaskDefinition.COMMAND_TASK:
	  getCommandButton() . setState ( true );
	  break;
	case DataTaskDefinition.EXCEPTION_TASK:
	  getExceptionButton() . setState ( true );
	  break;
	case DataTaskDefinition.MONITOR_TASK:
	  getMonitorButton() . setState ( true );
	  break;
	default:
	  System.err.println ( "[NewTaskWindowHandler:update]  Warning:  "
			       + "Unknown Task type:  "
			       + dataTaskDef . getTaskType() );
	  break;
      }

      getExternCheckbox()
	. setState ( dataTaskDef . getIsExtern() );
      getPlanFirstCheckbox() . setState (
	   dataTaskDef . getHasConstraint ( DataConstraint.EXPAND_FIRST ) );
      getDelayPlanningCheckbox() . setState (
	   dataTaskDef . getHasConstraint ( DataConstraint.DELAY_EXPANSION ) );

	/* Disable the Constraints if we are "extern" */
      getPlanFirstCheckbox()
	. setEnabled ( dataTaskDef . getIsExtern() == false );
      getDelayPlanningCheckbox()
	. setEnabled ( dataTaskDef . getIsExtern() == false );


      
	/*DBA-Alert:  How to update the locationFileTextField?
	 *            The information does not come from the DataComponent...
	 *            (getLocationDataFile is NULL for task-editing...)
	 */
      if ( getLocationDataFile() != null )
      {
	setTextFieldText ( getLocationFileTextField(),
			   getLocationDataFile() . getFilename() );
      }
      else
      {
	setTextFieldText ( getLocationFileTextField(), "" );
      }

      getTaskNameTextField() . setStringValue ( dataTaskDef . getTaskName() );


	/* Update the arguments */
      for ( i=0;  i < dataTaskDef . getTaskArgumentCount();  i++ )
      {
	  /* Are we reusing an argumentView? */
	if ( i < getArgumentSubviews() . count() )
	{
	  argumentObservableData = dataTaskDef . getTaskArgument ( i )
					       . findObserverData ( this );

	    /* Make certain that we are observing the arguments properly... */
	  if ( argumentObservableData == null )
	    dataTaskDef . getTaskArgument ( i )
                        . addObserver ( this , getArgumentSubview ( i ) );
	  else
	    argumentObservableData . setData ( getArgumentSubview ( i ) );


	    /* And lets use recursion to update this portion of ourself. */
	  update ( dataTaskDef . getTaskArgument ( i ),
		   dataTaskDef . getTaskArgument ( i ),
		   getArgumentSubview ( i ) );
	}
	else  /* Otherwise, just add a new argument normally */
	{
	  addAdditionalArgument ( dataTaskDef . getTaskArgument ( i ) );
	}
      }

	/* Remove any extra argument views */
      while ( i < getArgumentSubviews() . count() )
      {
	removeLastArgumentSubview ();
      }

	/* Note that we have succeeded */
      return true;

    } /* if ( theObservedComponent == getDataTaskDefinition() ) */



    System.err.println ( "[NewTaskWindowHandler:update]  Error:  Unable to "
			 + "update from unknown Observed Component (\""
			 + theObservedComponent . getClass() . getName()
			 + "\")." );
    return false;

  } /* public boolean update ( ... )  [DataComponentObserver Interface] */



	/* ExtendedTarget Interface */
  public boolean canPerformCommand ( String theCommand ) 
  {
    return 
      (    theCommand . equals( NewTaskWindowHandler.OK_COMMAND            )
	|| theCommand . equals( NewTaskWindowHandler.CANCEL_COMMAND        )
	|| theCommand . equals( NewTaskWindowHandler.ADD_ARGUMENTS_COMMAND )
	|| theCommand . equals( NewTaskWindowHandler.DATA_CHANGED_COMMAND  )
	|| theCommand . equals( NewTaskWindowHandler.SELECT_FILE_COMMAND   )
	|| theCommand . equals( NewTaskWindowHandler.
					      TOGGLE_IN_OUT_BUTTON_COMMAND ) );
  }

	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
    boolean   textParseDidFail = false;


	/* Ignore any pending events after, or while, our window goes away. */
    if ( getWindow() . isVisible() == false )
      return;


	/* If we are quiting/aborting, just quit... */
    if ( theCommand . equals( NewTaskWindowHandler.CANCEL_COMMAND ) )
    {
	/* Try to unregister us, so we don't create a spurious task... */
      if ( getDataTaskDefinition() != null )
      {
	getDataTaskDefinition() . unregisterTask (
				       "NewTaskWindowHandler:performCommand" );
      }
      getWindow() . hide();
      return;
    }


	/* If one of our TextFields or the plainCodeScrollText
	 * is currently focused and has changed data, bad things might happen.
	 * So lets terminate any text editing
	 */
    getRootView() . setFocusedView ( null );

	/* And abort this event if the plainCodeScrollText */
	/* is in the process of redisplaying itself...         */
    if (   ( getPlainCodeScrollText() . getDelayedAlertCount()       > 0 )
	|| ( getPlainCodeScrollText() . getTextView()
			              . getPendingSelectRangeCount() > 0 ) )
    {
	/* If we are aborting this event,
	 *  odds are one of our GUI widgets now has an incorrect state...
	 * So try to do a self-update to correct this misconception.
	 */
      if ( getDataTaskDefinition() != null )
	update ( getDataTaskDefinition(), getDataTaskDefinition(), null );
      return;
    }



	/* Handle successful conclusion. */
    if ( theCommand . equals ( NewTaskWindowHandler.OK_COMMAND ) )
    {
	/* If we are not currently editing anything... */
	/* (Then it's safe to exit...) */
      if ( getRootView().focusedView() == null )
      {
	if ( getDataTaskDefinition() . isValid() == false )
	{
	  runAlert ( "Error:", "Invalid Task Definition.",
		     "OK", null, null );
	}
	if ( getLocationDataFile() == null )
	{
	  runAlert ( "Error:",
		     "Invalid File Location:\n"
		     + "You must specify a file for this task.",
		     "OK", null, null );
	}
	else
	{
		/* Create a task-body, if necessary... */
	  if (   ( getDataTaskDefinition() . getIsExtern() == false )
	      && ( getDataTaskDefinition() . getTaskBody() == null  ) )
	  {
	    getDataTaskDefinition()
	      . setTaskBody ( new DataCompoundStatement ( "" ) );
      	  }

		/* Append this task to its DataFile... */
	  getLocationDataFile() . addSubcomponent ( "\n" );
	  getLocationDataFile() . addSubcomponent ( getDataTaskDefinition() );
	  getLocationDataFile() . addSubcomponent ( "\n" );

		/* Create a new window, if desired... */
	  String  openWindowCommand = null;

	  if ( getInternalWindowButton() . state() )
	    openWindowCommand = VDT.NEW_INTERNAL_WINDOW;

	  else if ( getExternalWindowButton() . state() )
	    openWindowCommand = VDT.NEW_EXTERNAL_WINDOW;

	  if ( openWindowCommand != null )
	  {
	    Target  target
	              = TargetChain . applicationChain()
				    . targetForCommand ( openWindowCommand );

	    if ( target != null )
	    {
	      Application . application()
		          . performCommandLater ( target,
						  openWindowCommand,
						  getDataTaskDefinition() );
	    }
	    else
	    {
	      System.err.println ( "[NewTaskWindowHandler:performCommand(OK)]"
		 + "  Error: Unable to find TARGET for opening New Window. (\""
		 + openWindowCommand + "\")" );
	    }
	  }

		/* Close this window... */
	  getWindow() . hide();

	} /* If getDataTaskDefinition() was valid */
      } /* If it is safe to exit... (Ie: Not currently editing anything) */
    } /* If ( theCommand . equals ( OK_CMMAND ) ) */

	/* Handle case of adding more arguments */
    else if ( theCommand
	      . equals( NewTaskWindowHandler.ADD_ARGUMENTS_COMMAND ) )
    {
      addAdditionalArgument ( null );
    }

	/* Handle case of data changing -- ie: regenerate text */
    else if ( theCommand
	      . equals( NewTaskWindowHandler.DATA_CHANGED_COMMAND ) )
    {
      performDataChangedCommand ( theData );
    }

	/* Handle case of selecting a file... */
    else if ( theCommand . equals( NewTaskWindowHandler.SELECT_FILE_COMMAND ) )
    {
      SelectFileWindowHandler.getSelectFileWindowHandler()
	. runModal ( getRootView() );

      if ( SelectFileWindowHandler.getSelectFileWindowHandler()
	     . getSelectedDataFile() != null )
      {
	setLocationDataFile (
		     SelectFileWindowHandler.getSelectFileWindowHandler()
		       . getSelectedDataFile() );
      }
    }

	/* Handle case of button needing to be toggled. */
    else if ( theCommand
	       . equals ( NewTaskWindowHandler.TOGGLE_IN_OUT_BUTTON_COMMAND ) )
    {
      performToggleInOutButtonCommand ( theData );
    }

    else
    {
      super . performCommand ( theCommand, theData );
    }

  } /* public void performCommand ( String theCommand, Object theData ) */



  public void performDataChangedCommand ( Object theData )
  {
    int                        i;
    boolean                    revertComponent;
    NewTaskWindowArgumentView  argumentsSubview;
    DataTaskArgument           dataTaskArgument;
    DataComponent              dataComponentToUpdate = getDataTaskDefinition();


	/* Abort if we are not part of the applications view hierarchy... */
    if ( getWindow() . isVisible() == false )
      return;


    if ( theData == getExternCheckbox() )
    {
      getDataTaskDefinition() . setIsExtern ( getExternCheckbox() . state());

	/* Disable the Constraints if we are "extern" */
      getPlanFirstCheckbox()     . setEnabled ( getExternCheckbox() . state()
						== false );
      getDelayPlanningCheckbox() . setEnabled ( getExternCheckbox() . state()
						== false );
    }

    else if ( theData == getTaskButton() )
      getDataTaskDefinition()
	. setTaskType ( DataTaskDefinition.TASK_TASK );

    else if ( theData == getGoalButton() )
      getDataTaskDefinition()
	. setTaskType ( DataTaskDefinition.GOAL_TASK );

    else if ( theData == getCommandButton() )
      getDataTaskDefinition()
	. setTaskType ( DataTaskDefinition.COMMAND_TASK );

    else if ( theData == getExceptionButton() )
      getDataTaskDefinition()
	. setTaskType ( DataTaskDefinition.EXCEPTION_TASK );

    else if ( theData == getMonitorButton() )
      getDataTaskDefinition()
	. setTaskType ( DataTaskDefinition.MONITOR_TASK );


    else if ( theData == getPlanFirstCheckbox() )
    {
      if (   getDataTaskDefinition()
	       . getHasConstraint ( DataConstraint.EXPAND_FIRST )
	   != getPlanFirstCheckbox() . state() )
      {
	if ( getPlanFirstCheckbox() . state() )
	{
	  DataConstraint  newConstraint = new DataConstraint ();
	  newConstraint . addSubcomponent   ( " "                          );
	  newConstraint . setConstraintType ( DataConstraint.EXPAND_FIRST  );
	  newConstraint . setIndex          ( DataConstraint.KEYWORD + 0   );
	  getDataTaskDefinition() . addConstraint ( newConstraint );

	  if ( getDelayPlanningCheckbox() . state() )
	  {
	    getDataTaskDefinition() . removeConstraint (
		  getDataTaskDefinition() . findConstraintIndex (
					    DataConstraint.DELAY_EXPANSION ) );
	    getDelayPlanningCheckbox() . setState ( false );
	  }
	}
	else
	{
	  getDataTaskDefinition() . removeConstraint (
		getDataTaskDefinition() . findConstraintIndex (
					       DataConstraint.EXPAND_FIRST ) );
	}
      }
    }

    else if ( theData == getDelayPlanningCheckbox() )
    {
      if (    getDataTaskDefinition()
	        . getHasConstraint ( DataConstraint.DELAY_EXPANSION )
	   != getDelayPlanningCheckbox() . state() )
      {
	if ( getDelayPlanningCheckbox() . state() )
	{
	  DataConstraint  newConstraint = new DataConstraint ();
	  newConstraint . addSubcomponent   ( " "                            );
	  newConstraint . setConstraintType ( DataConstraint.DELAY_EXPANSION );
	  newConstraint . setIndex          ( DataConstraint.KEYWORD + 0     );
	  getDataTaskDefinition() . addConstraint ( newConstraint );

	  if ( getPlanFirstCheckbox() . state() )
	  {
	    getDataTaskDefinition() . removeConstraint (
		  getDataTaskDefinition() . findConstraintIndex (
					       DataConstraint.EXPAND_FIRST ) );
	    getPlanFirstCheckbox() . setState ( false );
	  }
	}
	else
	{
	  getDataTaskDefinition() . removeConstraint (
		getDataTaskDefinition() . findConstraintIndex (
					    DataConstraint.DELAY_EXPANSION ) );
	}
      }
    }

	/* Deal with file-location... */
    else if ( theData == getLocationFileTextField() )
    {
      if (   (   ( getLocationDataFile() == null )
	      && ( getLocationFileTextField() . stringValue() . length() > 0 )
	      )
	  || (   ( getLocationDataFile() != null )
	      && ( getLocationFileTextField() . stringValue() . equals (
					getLocationDataFile() . getFilename() )
		   == false )
	      )
	  )
      {
	DataFile  newDataFile
		    = WindowHandlerAbstractBaseClass.validateNewFilename (
				  getLocationFileTextField() . stringValue(),
				  getRootView(), true );

	if ( newDataFile != null )
	{
	  setLocationDataFile ( newDataFile );
	}

	else if ( getLocationDataFile() != null )
	{
	  setTextFieldText ( getLocationFileTextField(),
			     getLocationDataFile() . getFilename() );
	}
	else
	{
          setTextFieldText ( getLocationFileTextField(), "" );
	}
      }

	/* Location is not part of the task-data-component. */
      dataComponentToUpdate = null;
    }


    else if ( theData == getTaskNameTextField() )
    {
      if ( /*NOT*/ !  ( getTaskNameTextField() . stringValue()
			. equals( getDataTaskDefinition() . getTaskName() ) ) )
      {
	revertComponent = true;
	try
	{
	  if ( getDataTaskDefinition() . setTaskName (
			    getTaskNameTextField() . stringValue() ) == false )
	  {
	    if ( ScrollTextOfDataComponent.showReeditRevertAlert (
			   "Internal Error:", "Unable to set Task name.",
			   getRootView() ) )
	    {
	      Application . application() . performCommandLater (
		 getTaskNameTextField(), TextField.SELECT_TEXT, null, false );
	      revertComponent = false;
	    }
	    dataComponentToUpdate = null;
	  }
	}
	catch ( DetailedParseException  theException )
	{
	  if ( ScrollTextOfDataComponent.showReeditRevertAlert (
					   "Error - Invalid Task Name:",
					   theException . toClippedString(),
					   getRootView() ) )
	  {
	    Application . application() . performCommandLater (
		 getTaskNameTextField(), TextField.SELECT_TEXT, null, false );
	    revertComponent = false;
	  }
	  dataComponentToUpdate = null;
	}
	if ( revertComponent ) /* Set worked or Revert'ing */
	{
	  setTextFieldText ( getTaskNameTextField(),
			     getDataTaskDefinition() . getTaskName() );
	}
      }
      else /* TextField text has not changed. */
      {
	dataComponentToUpdate = null;
      }
    }

	/* Are we updating arguments? */
    else if (   ( theData  instanceof View )
	     && ( ( ((View) theData) . superview() )
		  instanceof NewTaskWindowArgumentView ) )
    {
      argumentsSubview
	= (NewTaskWindowArgumentView) ( ((View) theData) . superview() );

	/* Warn us about potential problems */
      if (  getArgumentSubviews() . contains ( argumentsSubview ) == false )
      {
	System.err.println ( "NewTaskWindowHandler:performDataChangedCommand] "
			     + " Warning:  theData . superview() is of type "
			     + "NewTaskWindowArgumentView, but is not a known "
			     + "NewTaskWindowArgumentView...  "
			     + "Trying to continue..." );
      }

	/* Find the underlying DataComponent object */
      dataTaskArgument = findDataTaskArgumentForView ( argumentsSubview );
      dataComponentToUpdate = dataTaskArgument;


      if ( dataTaskArgument == null )
      {
	System.err.println ( "NewTaskWindowHandler:performDataChangedCommand] "
			     + " Error:  theData . superview() is of type "
			     + "NewTaskWindowArgumentView, but it has NO "
			     + "KNOWN corresponding DataTaskArgument.   "
			     + "Unable to find/update underlying DataComponent"
			     + " object.  [\"" + theData + "\"]" );
	dataComponentToUpdate = null;
      }
      else
      {
	if ( theData == argumentsSubview . getArgument() )
	{
	  if ( /*NOT*/ ! ( ( (TextField) theData ) . stringValue()
			   . equals ( dataTaskArgument . toString() ) ) )
	  {
	    revertComponent = true;
	    try
	    {
	      if ( dataTaskArgument . parse (
			   ( (TextField) theData ) . stringValue(),
			   DataTaskArgument.TYPE_NAME_AND_EQUALS   ) == false )
	      {
		throw new DetailedParseException (
			      "Failure in DataTaskArgument Parsing routine." );
	      }
	    }
	    catch ( DetailedParseException  theException )
	    {
	      if ( ScrollTextOfDataComponent.showReeditRevertAlert (
			 "Error:",
			 "Invalid Argument:\n\n"
			 + theException.toClippedString(),
			 getRootView() ) )
	      {
		Application . application() . performCommandLater (
		     ((Target) theData), TextField.SELECT_TEXT, null, true );
		revertComponent = false;
	      }
	      dataComponentToUpdate = null;
	    }
	    if ( revertComponent ) /* Set worked or Revert'ing */
	    {
	      setTextFieldText ( 
			(TextField) theData,
			dataTaskArgument
			  . toString( DataTaskArgument.TYPE_NAME_AND_EQUALS) );
	    }
	  }
	  else /* TextField text has not changed. */
	  {
	    dataComponentToUpdate = null;
	  }
	}

	else
	{
	  System.err.println( "NewTaskWindowHandler:performDataChangedCommand]"
			     + "  Error:  theData is a subview of a KNOWN "
			     + "NewTaskWindowArgumentView, but it is NOT the "
			     + "argument-type or argument-name TextField.  "
			     + "There is NO KNOWN counterpart to update...  "
			     + "Unable to find/update underlying DataComponent"
			     + " object.  [\"" + theData + "\"]" );

	  dataComponentToUpdate = null;
	}
      }
    }
    else
    {
      System.err.println ( "NewTaskWindowHandler:performDataChangedCommand] "
			   + " Warning:  DATA_CHANGED_COMMAND for unknown "
			   + "target:  " + theData );
      dataComponentToUpdate = null;
    }

	/* In any case, if we updated the DataTaskDefinition object,    */
	/* we need to notify any other observers that it has changed... */
    if ( dataComponentToUpdate != null )
      dataComponentToUpdate . update ( this );

  } /* public void performDataChangedCommand ( Object theData ) */


  public void performToggleInOutButtonCommand ( Object theData )
  {
	/* Convenience Variables */
    Button            theButton;
    DataTaskArgument  dataTaskArgument;
    int               i;

    if ( /*NOT*/ !  ( theData instanceof Button ) )
    {
      System.err.println ( "[NewTaskWindowHandler:"
	     + "performToggleInOutButtonCommand]  Error:  "
	     + "Invalid Argument Class Type.  Expected a Button!  (\"" 
	     + theData . getClass() . getName() + "\")" );
      return;
    }

	/* Convenience variable */
    theButton = (Button) theData;

	/* Handle known-state case */
    for ( i=0; i < NewTaskWindowHandler.INOUT_BUTTON_TITLES.length; i++ )
    {
      if ( theButton . title()
	   . equals ( NewTaskWindowHandler.INOUT_BUTTON_TITLES [ i ] ) )
      {
	theButton . setTitle ( NewTaskWindowHandler.INOUT_BUTTON_TITLES [ 
	       ( i + 1 ) % NewTaskWindowHandler.INOUT_BUTTON_TITLES.length ] );
	break; /* Exit For Loop */
      }
    }

	/* IF ( we didn't break out of the loop) */
    if ( i >= NewTaskWindowHandler.INOUT_BUTTON_TITLES.length )
    {
	/* Handle unknown-state case */
      System.err.println ( "[NewTaskWindowHandler:"
	     + "performToggleInOutButtonCommand]  Warning:  "
	     + "Unknown Button state:  \"" + theButton . title() + "\"" );

      theButton . setTitle( NewTaskWindowHandler.INOUT_BUTTON_TITLES [ 0 ] );
    }

	/* Update the dataTaskArgument... */
    dataTaskArgument = findDataTaskArgumentForView ( theButton.superview() );
    if ( dataTaskArgument == null )
    {
      System.err.println ( "[NewTaskWindowHandler:"
	     + "performToggleInOutButtonCommand]  Warning:  "
	     + "No known DataTaskArgument found for this "
	     + "Button's superview.  Unable to find/update underlying "
	     + "DataComponent object.  [\"" + theButton . title() + "\"]" );
    }
    else
    {
      dataTaskArgument . setArgumentDirectionString ( theButton . title() );
      dataTaskArgument . update ( this );
    }

  } /* public void performToggleInOutButtonCommand ( Object theData ) */




	/* TextFieldOwner Interface */
  public void    textEditingDidBegin ( TextField theTextField ) { }
  public boolean textEditingWillEnd  ( TextField theTextField,
				       int       theEndCondition,
				       boolean   theContentsChanged )
							       { return true; }
  public void    textEditingDidEnd   ( TextField theTextField,
				       int       theEndCondition,
				       boolean   theContentsChanged )
  {
	/* Ignore any pending events after, or while, our window goes away. */
    if ( getWindow() . isVisible() == false )
      return;
    else
      performDataChangedCommand ( theTextField );
  }

  public void textWasModified ( TextField theTextField )
  { }


	/* WindowOwner Interface */
  public void windowDidBecomeMain ( Window theWindow ) {}
  public void windowDidResignMain ( Window theWindow ) {}
  public void windowDidShow       ( Window theWindow ) {}
  public boolean windowWillShow   ( Window theWindow ) { return true; }
  public void windowWillSizeBy    ( Window theWindow, Size theDeltaSize ) {}

  public boolean windowWillHide   ( Window theWindow )
  {
    return true;
  }

  public void windowDidHide       ( Window theWindow )
  {
	/* We can go away now...  Lets stop watching everything... */
    clearDataTaskDefinition();
  }

}
