
import netscape.application.*;

public class SimpleLabel extends View implements Target, ExtendedTarget
{
	/* Note:  use delayedRedraw() to activate this target */
  public  final static String DELAYED_REDRAW = "DELAYED_REDRAW";

  public final static int  DEFAULT_MINIMUM_WIDTH
		  = Font.defaultFont().fontMetrics().stringWidth  ( "A" );
  public final static int  DEFAULT_MINIMUM_HEIGHT
		  = Font.defaultFont().fontMetrics().stringHeight (     );


  protected Color   foregroundColor, backgroundColor;
  protected String  text;
  protected Font    font;
  protected boolean doingDelayedRedraw;


  public SimpleLabel()  { this ( "", null ); }

  public SimpleLabel ( String theText )  { this ( theText, null ); }

  public SimpleLabel ( String theText, Font theFont )
  {
    foregroundColor = Color.black;
    backgroundColor = Color.white;
    text = null;
    font = null;
    doingDelayedRedraw = false;
    if ( theFont != null )
      setFont ( theFont );
    setText ( theText );
    setBuffered ( false );
  }

  public boolean isTransparent ( ) { return false; }


  public Color getForegroundColor ( ) { return foregroundColor; }
  public void  setForegroundColor ( Color theForegroundColor )
				      { foregroundColor = theForegroundColor; }

  public Color getBackgroundColor ( ) { return backgroundColor; }
  public void  setBackgroundColor ( Color theBackgroundColor )
				      { backgroundColor = theBackgroundColor; }

  public String getText ( ) { return text; }
  public void   setText ( String theText )
  {
    if ( theText == null )
      text = "";
    else
      text = theText;

    Size size = getFont() . fontMetrics() . stringSize ( getText() );

	/* Make certain that we have at least *some* room */
	/* to display *something*!                        */
    if ( size . height < SimpleLabel . DEFAULT_MINIMUM_HEIGHT )
      size . height = SimpleLabel . DEFAULT_MINIMUM_HEIGHT;

	/* Allow a little extra border-spacing */
    size . width  += SimpleLabel . DEFAULT_MINIMUM_WIDTH;
    size . height += 3; /* Random magic number */

    sizeTo ( size . width,  size . height );
  }


  public Font getFont ( )
  {
    return font != null ? font : Font.defaultFont();
  }

  public void setFont ( Font theFont )
  {
    font = theFont;
    setText ( getText() );
  }


  public void drawView ( Graphics theGraphics )
  {
    if ( getBackgroundColor() != null )
    {
      theGraphics . setColor ( getBackgroundColor() );
      theGraphics . fillRect ( 0, 0, bounds().width, bounds().height );
    }
    theGraphics . setColor ( getForegroundColor() );
    theGraphics . setFont  ( getFont() );

    Size  size = getFont() . fontMetrics() . stringSize ( getText() );
    theGraphics . drawString ( getText(), 
			       ( bounds() . width  -  size . width ) / 2,
			       ( bounds() . height 
				 - 1  /* Random magic number */
				 - ( ( bounds() . height - size . height )
				     / 2 )
				 - ( getFont() . fontMetrics() . descent() ) )
			      );

	/* Hack to solve TextViews-turning-Black bug.
	 * Do a single delayed-redraw after every normal drawView...
	 */
    if ( doingDelayedRedraw == true )
    {
      doingDelayedRedraw = false;
    }
    else
      delayedRedraw();
  }


  public void delayedRedraw()
  {
    Application . application()
                . performCommandLater ( this, SimpleLabel.DELAYED_REDRAW,
					null, true );
  }

  public boolean canPerformCommand ( String theCommand )
  {
    return theCommand . equals ( SimpleLabel.DELAYED_REDRAW );
  }

  public void performCommand ( String theCommand, Object theData )
  {
    if ( theCommand . equals ( SimpleLabel.DELAYED_REDRAW ) )
    {
      doingDelayedRedraw = true;
      setDirty ( true );
    }
    else
    {
      System.err.println ( "\n[SimpleLabel:performCommand]  Warning:  "
			   + "Unknown command \"" + theCommand + "\"\n" );
    }
  }

}
