
import netscape.application.*;
import netscape.util.*;
import netscape.constructor.*;
import java.io.IOException;
import java.io.File;

/*Abstract class*/
public class WindowHandlerAbstractBaseClass extends Object
                                         implements Target
{
	/* Constants */
  public final static int    WINDOW_X_PADDING        = 5;
  public final static int    WINDOW_Y_PADDING        = 5;
  public final static int    WINDOW_WIDTH_PADDING    = 10;
  public final static int    WINDOW_HEIGHT_PADDING   = 15;
  public final static String FILEPATH_SEPARATOR 
				   = System.getProperty("file.separator", "/");




	/* Class Variables */
  protected static String    priorSelectedDirectory  = null;


	/* Class Methods */
  protected static synchronized Plan getPlan ( String thePlanName,
					       String theErrorMessage )
  {
    try
    {
      return
	new Plan ( Parameters.getParameters().getPlanPath ( thePlanName ) );
    }
    catch ( IOException theException )
    {
      System.err.println ( theErrorMessage + " [\"" + thePlanName
			   + "\"] Plan.   (" + theException + ")" );
      return null;
    }
  }

  public static String getPriorSelectedDirectory()
  {
    if ( WindowHandlerAbstractBaseClass.priorSelectedDirectory != null )
    {
      return WindowHandlerAbstractBaseClass.priorSelectedDirectory;
    }
    else
    {
      return System.getProperty ( "user.dir" );
    }
  }

  public static void   setPriorSelectedDirectory(
					     String thePriorSelectedDirectory )
  {
    WindowHandlerAbstractBaseClass.priorSelectedDirectory
      = thePriorSelectedDirectory;
  }


  public static DataFile validateNewFilename ( String    theNewFilename,
					       RootView  theRootView,
					       boolean   theCanCreate )
  {
    DataFile  dataFile;
    File      newFile;


	/* Do we already have this file loaded in? */
    if ( DataFile.getIsFilenameRegistered ( theNewFilename ) )
    {
      dataFile = DataFile.getFileForPath ( theNewFilename );

      if ( dataFile == null )
      {
	System.err.println (
		       "[WindowHandlerAbstractBaseClass:validateNewFilename]  "
		       + "Unable to locate registered file \""
		       + theNewFilename + "\"" );
      }

      return dataFile;
    }


    newFile = new File ( theNewFilename );

	/* Does this new file exists & is a directory? */
    if ( newFile . isDirectory() )
    {
      UTIL.runAlert ( "File Selection Error:",
		      "File \"" + theNewFilename + "\" is a directory.",
		      "Abort", null, null, theRootView );
      return null;
    }

	/*Does this file already exist? (As something other than a directory)*/
    if ( newFile . exists() )
    {
      try
      {
	return DataFile . loadDataFile ( newFile );
      }
      catch ( DetailedParseException theException )
      {
	UTIL.runAlert ( "File Selection Error:",
		      "File \"" + theNewFilename + "\" already exists\n"
		      + "and can not be parsed as TDL code.\n\n"
		      + "Parse Failure:\n" + theException . toClippedString(),
		      "Abort", null, null, theRootView );
      }
      return null;
    }

    else if ( theCanCreate )  /* And newFile . exists() == false */
    {
	/* Or is it a new file? */
      DataFile  newDataFile = new DataFile ( theNewFilename );
      newDataFile
	. registerFile( "WindowHandlerAbstractBaseClass:validateNewFilename" );
      return newDataFile;
    }

    else /* newFile not dir, not exist, theCanCreate == false */
    {
      UTIL.runAlert ( "File Selection Error:",
		      "File \"" + theNewFilename + "\" does NOT exist.",
		      "Abort", null, null, theRootView );
      return null;
    }
  }



  public static String  resolveFilepath ( String    theFilepath,
					  String    theDefaultFilepath,
					  String    theErrorMessage,
					  RootView  theRootView )
  {
	/* If there is a FILEPATH_SEPARATOR in theFilepath,      */
	/* Assume that it is a absolute path, and we are done... */
    if ( theFilepath . indexOf ( FILEPATH_SEPARATOR )  !=  -1 )
    {
      return theFilepath;
    }

    /* Else assume that it is a relative path, off theDefaultFilepath... */

	/* Do we have a default filepath? */
    if (   ( theDefaultFilepath            == null                         )
	|| ( theDefaultFilepath . length() <= 0                            )
	|| ( theDefaultFilepath . lastIndexOf ( FILEPATH_SEPARATOR ) == -1 ) )
    {
      if (   ( theErrorMessage == null )
	  || ( UTIL.runAlert ( "Warning:",
			       "No relative path/directory available for "
			       + theErrorMessage
			       + " \""
			       + FILEPATH_SEPARATOR + theFilepath +"\"?",
			       "Continue", "Abort", null, theRootView )
	       != Alert.DEFAULT_OPTION ) )
      {
	return null;
      }
      else
      {
	return FILEPATH_SEPARATOR + theFilepath;
      }
    }

    else /* We do have a base-path */

    {
      File  tmpFile = new File ( theDefaultFilepath );

	/* Should we consider theDefaultFilepath to be a directory? */
      if (   ( tmpFile . isDirectory() == true                             )
	  || (   ( tmpFile . exists()  == false                          )
	      && (  theDefaultFilepath . endsWith ( FILEPATH_SEPARATOR ) ) ) )
      {
	if ( theDefaultFilepath . endsWith ( FILEPATH_SEPARATOR ) == false )
	{
	  theDefaultFilepath = theDefaultFilepath + FILEPATH_SEPARATOR;
	}
	/* Else we are already all set up... */
      }
      else /* Should we consider theDefaultFilepath to be a file? */
      {
	theDefaultFilepath
	  = theDefaultFilepath . substring (
		        0, 
			theDefaultFilepath . lastIndexOf ( FILEPATH_SEPARATOR )
			+ FILEPATH_SEPARATOR . length() );
      }

      return theDefaultFilepath + theFilepath;
    }

  }


	/* Bug-work-around for TextField. */
  public static void setTextFieldText ( TextField theTextField,
					String    theString    )
  {	/* There is a bug in TextField when the strings match... */
    if ( /*NOT*/ ! ( theTextField . stringValue() . equals ( theString ) ) )
    {
      theTextField . setStringValue ( theString );
    }
  }






	/* Instance Variables */
  protected Window  window;
  protected VDT     vdt;

  
	/* Instance Methods */
  public WindowHandlerAbstractBaseClass()
  {
    window = null;
    vdt    = null;
  }

	/* Constructors can't be synchronized. */
	/* And this needs to be synchronized since we use the static plans */
  protected synchronized
      boolean createAndInitialize ( Plan     thePlan,
				    boolean  theUseExternalWindow,
				    String   theWindowTitle,
				    String   theErrorMessageLocation,
				    VDT      theVDT )
  {
	/* Do we have a Plan object? */
    if ( thePlan == null )
    {
      System.err.println (
		  "[WindowHandlerAbstractBaseClass:createAndInitialize("
		  + theErrorMessageLocation
		  + ")]  Error:  Null Plan.  Failing..." );
      return false;
    }

	/* Create a new set of objects from the plan */
    if ( thePlan . unarchiveObjects ( this ) == false )
    {
      System.err.println (
		  "[WindowHandlerAbstractBaseClass:createAndInitialize("
		  + theErrorMessageLocation
		  + ")]  Warning:  thePlan.unarchiveObjects() "
		  + "returned false.  Possible failure." );
    }

	/* resize our plan */
    thePlan.sizeToFit();
    thePlan.moveBy  ( WindowHandlerAbstractBaseClass.WINDOW_X_PADDING,
		      WindowHandlerAbstractBaseClass.WINDOW_Y_PADDING );
    thePlan.setSize ( thePlan . size() . width
		      + WindowHandlerAbstractBaseClass.WINDOW_WIDTH_PADDING,
		      thePlan . size() . height
		      + WindowHandlerAbstractBaseClass.WINDOW_HEIGHT_PADDING );

	/* Create our window */
    if ( theUseExternalWindow )
    {
      window = thePlan . externalWindowWithContents();

	/* If we are basing this window off the VDT object... */
      if ( theVDT != null )
      {
	Menu   newMenu = new Menu();

	theVDT . copyMenu ( newMenu, (ExternalWindow) window );
	((ExternalWindow) window) . setMenu ( newMenu );
      }
    }
    else
    {
      window = thePlan . internalWindowWithContents();
      ( (InternalWindow) window ) . setCloseable ( true );
    }

	/* Configure our window */
    window . setMinSize ( window . bounds() . width,
			  window . bounds() . height );
    window . setResizable ( true );
	/* If we are basing this window off the VDT object... */
    if ( theVDT != null )
    {
      window . setTitle  ( theWindowTitle );
      window . setOwner  ( theVDT );
      theVDT . addWindow ( window );
    }

	/* Initialize vdt (cache theVDT) */
    vdt = theVDT;


    return true;
  }


  public void runModal ( )
  {
	/* Show the window modally */
    getWindow() . showModally();
  }


  public void runNoModal ( )
  {
	/* Show the window */
    getWindow() . show();
  }


  public void run ( )
  {
    runNoModal();
  }


  public void runModal ( RootView theRootView )
  {
    if ( getIsInternalWindow() )
    {
      if ( theRootView != null )
	((InternalWindow) getWindow()) . setRootView ( theRootView );
      else
	((InternalWindow) getWindow())
	  . setRootView ( Application.application() . mainRootView() );
    }

    runModal ();
  }


  public void run ( RootView theRootView )
  {
    if ( getIsInternalWindow() )
    {
      if ( theRootView != null )
	((InternalWindow) getWindow()) . setRootView ( theRootView );
      else
	((InternalWindow) getWindow())
	  . setRootView ( Application.application() . mainRootView() );
    }

    run();
  }


  public int runAlert ( String theTitle,
			String theMessage,
			String theDefaultOption,
			String theSecondOption,
			String theThirdOption )
  {
    RootView   rootView;

    if ( getIsInternalWindow() )
      rootView = ((InternalWindow) getWindow()) . rootView();
    else
      rootView = ((ExternalWindow) getWindow()) . rootView();

    return UTIL.runAlert ( theTitle,         theMessage,
			   theDefaultOption, theSecondOption,
			   theThirdOption,   rootView );
  }

  public boolean getIsInternalWindow()
  {
    return getWindow() instanceof InternalWindow;
  }

  public boolean getIsExternalWindow()
  {
    return getWindow() instanceof ExternalWindow;
  }


  public Window getWindow() { return window; }
  public VDT    getVDT()    { return vdt; }


	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
    System.out.println ( "[WindowHandlerAbstractBaseClass("
			 + getClass() . getName()
			 + "):performCommand]  Unknown Command: \""
			 + theCommand + "\"" );
  }
}
