
/*
 * This class is utilized with respect to Distributed Tasks.
 * It allows task's to register themselves via the initialization a
 * static variable containing an instance of this class.
 * Later on, in TDL_INITIALIZE(), we exploit these registrations to
 * register these distributed-tasks appropriately with TCM.
 *
 * Note:  The second argument to the _TDL_DistributedRegistryEntry constructor
 * (theDistributedRemoteTaskName) is a default task-instance-name for the
 * remote task (e.g. the goal, command, monitor, etc).  This string overrides
 * the default value specified in the allocation function, AND IS IN TURN
 * OVERRIDEN BY the remote-task-instance name specified to the
 * TCM_AllocateDistributedNode() function!
 */

#define _TDL_INTERNAL_
#include "tdl.H"
#include "_TDL_Snode.H"
#include "_TDL_Slist.H"

#ifndef _TDL_DISTRIBUTED_REGISTRY_ENTRY
#define _TDL_DISTRIBUTED_REGISTRY_ENTRY

typedef TCM_Task_Tree_Ref ( * _TDL_DistributedAllocationFunctionType ) 
				 ( STRING, STRING, BOOLEAN );

typedef TCM_Action_Ref ( * _TDL_DistributedCreateRemoteActionFunctionType ) 
				 ( const TCM_Task_Tree_Ref &, void *, void * );

class _TDL_DistributedRegistryEntry : _TDL_Snode
{
	/* Class methods */
protected:
  static _TDL_Slist & getAllDistributedRegistriesList();

  static status_t addRegistryToMasterList (
		 _TDL_DistributedRegistryEntry * theDistributedRegistryEntry )
    { return _TDL_DistributedRegistryEntry::getAllDistributedRegistriesList()
	       . appendNode ( theDistributedRegistryEntry ); }

public:
  static int4              registerAllDistributedRegistryEntries();
  static TCM_Task_Tree_Ref genericDistributedRemoteTaskAlloctionFunction ( 
						     void * theTcmClientData );



	/* Instance variables */
protected:
  STRING                                 taskName;
	/* See note at top about theDistributedRemoteTaskName. */
  STRING                                 distributedRemoteTaskName;

  _TDL_DistributedAllocationFunctionType distributedAllocationFunction;

  _TDL_DistributedCreateRemoteActionFunctionType 
					 distributedCreateRemoteActionFunction;

  STRING                                 distributedFormatString;
  BOOLEAN                                hasBeenRegisteredWithTcm;



	/* Instance methods */
public:
  _TDL_DistributedRegistryEntry (
    STRING                            theTaskName,
	/* See note at top about theDistributedRemoteTaskName. */
    STRING                            theDistributedRemoteTaskName,
    _TDL_DistributedAllocationFunctionType
				      theDistributedAllocationFunction,
    _TDL_DistributedCreateRemoteActionFunctionType
				      theDistributedCreateRemoteActionFunction,
    STRING                            theDistributedFormatString )

    : _TDL_Snode                    (                                        ),
      taskName                      ( theTaskName                            ),
      distributedRemoteTaskName     ( theDistributedRemoteTaskName           ),
      distributedAllocationFunction ( theDistributedAllocationFunction       ),
      distributedCreateRemoteActionFunction (
				    theDistributedCreateRemoteActionFunction ),
      distributedFormatString       ( (    theDistributedFormatString
					== STRING(NULL) )
				     ? STRING(NULL)
				     : NewString (theDistributedFormatString)),
      hasBeenRegisteredWithTcm      ( FALSE                                  )
    { _TDL_DistributedRegistryEntry::addRegistryToMasterList ( this ); }


  virtual ~_TDL_DistributedRegistryEntry();

  
  STRING getTaskName() const
    { return taskName; }

	/* See note at top about theDistributedRemoteTaskName. */
  STRING getDistributedRemoteTaskName() const
    { return distributedRemoteTaskName; }

  _TDL_DistributedAllocationFunctionType
  getDistributedAllocationFunction()     const
    { return distributedAllocationFunction; }

  _TDL_DistributedCreateRemoteActionFunctionType 
  getDistributedCreateRemoteActionFunction()     const
    { return distributedCreateRemoteActionFunction; }

  STRING getDistributedFormatString() const
    { return distributedFormatString; }

  BOOLEAN  getHasBeenRegisteredWithTcm() const
    { return hasBeenRegisteredWithTcm; }


	/* Abstract iostreamBase-interface method */
  virtual ostream & printObject ( ostream    & theOstream,
				  const char * theIndentString = "") const;

protected:
  void     setHasBeenRegisteredWithTcm()
    { hasBeenRegisteredWithTcm = TRUE; }


	/* These operations are disabled (prohibited) & unimplemented */
private:
  _TDL_DistributedRegistryEntry (
		       const _TDL_DistributedRegistryEntry & theObjectToCopy );
  _TDL_DistributedRegistryEntry & operator = (
		       const _TDL_DistributedRegistryEntry & theObjectToCopy );
};

#endif /* _TDL_DISTRIBUTED_REGISTRY_ENTRY */
